# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import asyncio
import datetime
import json
import logging
import pprint
import random
import string
import time
import weakref
from collections import defaultdict

import aiohttp
from yarl import URL

from abot.bot import Backend, BotObject, Channel, Entity, Event, MessageEvent

logger = logging.getLogger('abot.dubtrack')
logger_layer1 = logging.getLogger('abot.dubtrack.layer1')
logger_layer2 = logging.getLogger('abot.dubtrack.layer2')
logger_layer3 = logging.getLogger('abot.dubtrack.layer3')

logger.setLevel(logging.INFO)
logger_layer1.setLevel(logging.INFO)
logger_layer2.setLevel(logging.INFO)
logger_layer3.setLevel(logging.INFO)


# Dubtrack specific objects
class DubtrackObject(BotObject):
    def __init__(self, data, dubtrack_backend: 'DubtrackBotBackend'):
        self._data = data
        self._dubtrack_backend = dubtrack_backend

    @property
    def backend(self):
        return self._dubtrack_backend


class DubtrackChannel(DubtrackObject, Channel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dubtrack_backend._register_user(self._data)

    async def say(self, text: str):
        if not self._dubtrack_backend.dubtrack_id:
            raise ValueError('You need to login to speak')
        await self._dubtrack_backend.dubtrackws.say_in_room(text)

    @property
    def entities(self):
        entities = []
        for user in self._dubtrack_backend.dubtrack_users:
            entity = self._dubtrack_backend._get_entity(user)
            if entity:
                entities.append(entity)
        return entities

    def __repr__(self):
        cls = self.__class__.__name__
        name = self._data['name']
        id = self._data['_id']
        slug = self._data['roomUrl']
        entities = self.entities
        return f'<{cls} {slug}#{id} name="{name}" {entities}>'


class DubtrackEntity(DubtrackObject, Entity):
    async def tell(self, text: str):
        pass

    @property
    def username(self):
        return self._data.get('username')

    @property
    def id(self):
        return self._data.get('id')

    @property
    def dubs(self):
        return self._data.get('dubs')

    @property
    def played_count(self):
        return self._data.get('played_count')

    @property
    def skips(self):
        return self._data.get('skips')

    @property
    def songs_in_queue(self):
        return self._data.get('songs_in_queue')

    def __repr__(self):
        cls = self.__class__.__name__
        userid = self.id
        username = self.username
        dubs = self.dubs
        played = self.played_count
        skips = self.skips
        songs = self.songs_in_queue
        return f'<{cls} {username}#{userid} {dubs}/{played} {skips}(S) ##{songs}>'

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.id == other.id:
            return True
        return False


class DubtrackEvent(DubtrackObject, Event):
    _data_type = ''

    @classmethod
    def from_data(cls, data, dubtrack_backend: 'DubtrackBotBackend'):
        for cl in cls.__subclasses__():
            if cl._data_type == data['type']:
                return cl(data, dubtrack_backend)
        else:
            return cls(data, dubtrack_backend)

    @property
    def sender(self) -> DubtrackEntity:
        pass

    @property
    def channel(self) -> DubtrackChannel:
        # Return the channel used to send the Event
        if hasattr(self, '_channel'):
            return self._channel
        raise ValueError('Channel is not set')

    @channel.setter
    def channel(self, channel: DubtrackChannel):
        if hasattr(self, '_channel'):
            raise ValueError(f'Channel {self._channel} is in place, cannot replace with {channel}')
        self._channel = channel

    async def reply(self, text: str, to: str = None):
        # Reply to the message mentioning if possible
        return

    def __repr__(self):
        cls = self.__class__.__name__
        return f'<{cls} #{self._data["type"]}>'


class DubtrackMessage(DubtrackEvent, MessageEvent):
    _data_type = 'chat-message'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dubtrack_backend._register_user(self._data.get('user'))

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('user', {}).get('username'))

    @property
    def text(self):
        return self._data.get('message')

    @property
    def message_id(self):
        return self._data.get('chatid')

    def __repr__(self):
        cls = self.__class__.__name__
        chatid = self.message_id
        sender = self.sender
        msg = self.text
        return f'<{cls}#{chatid} {sender} "{msg}">'


class DubtrackSkip(DubtrackEvent):
    _data_type = 'chat-skip'

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('username'))

    def __repr__(self):
        cls = self.__class__.__name__
        sender = self.sender
        return f'<{cls} {sender}>'


class DubtrackDelete(DubtrackEvent):
    _data_type = 'delete-chat-message'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dubtrack_backend._register_user(self._data['user'])

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('user', {}).get('username'))

    @property
    def message_id(self):
        return self._data.get('chatid')

    def __repr__(self):
        cls = self.__class__.__name__
        chatid = self.message_id
        sender = self.sender
        return f'<{cls}#{chatid} {sender}>'


class DubtrackDub(DubtrackEvent):
    _data_type = 'room_playlist-dub'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dubtrack_backend._register_user(self._data['user'])

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('user', {}).get('username'))

    @property
    def dubtype(self):
        return self._data.get('dubtype')

    @property
    def total_updubs(self):
        return self._data.get('playlist', {}).get('updubs')

    @property
    def total_downdubs(self):
        return self._data.get('playlist', {}).get('downdubs')

    @property
    def length(self):
        song_length = self._data.get('playlist', {}).get('songLength')
        if song_length:
            return datetime.timedelta(milliseconds=song_length)

    @property
    def played(self):
        played = self._data.get('playlist', {}).get('played')
        if played:
            return datetime.datetime.utcfromtimestamp(played / 1000)

    def __repr__(self):
        cls = self.__class__.__name__
        dubtype = self.dubtype
        sender = self.sender
        downdubs = self.total_downdubs
        updubs = self.total_updubs
        return f'<{cls}#{dubtype} {sender} +{updubs}-{downdubs}>'


class DubtrackRoomQueueReorder(DubtrackEvent):
    _data_type = 'room_playlist-queue-reorder'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dubtrack_backend._register_user(self._data['user'])

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('user', {}).get('username'))

    def __repr__(self):
        cls = self.__class__.__name__
        sender = self.sender
        return f'<{cls} {sender}>'


class DubtrackUserQueueUpdate(DubtrackEvent):
    _data_type = 'room_playlist-queue-update-dub'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dubtrack_backend._register_user(self._data['user'])

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('user', {}).get('username'))

    def __repr__(self):
        cls = self.__class__.__name__
        sender = self.sender
        return f'<{cls} {sender}>'


class DubtrackPlaying(DubtrackEvent):
    _data_type = 'room_playlist-update'

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('song', {}).get('userid'))

    def __repr__(self):
        cls = self.__class__.__name__
        name = self.song_name
        songtype = self.song_type
        songfkid = self.song_external_id
        songid = self.song_id
        sender = self.sender
        return f'<{cls}#{songid} {songtype}#{songfkid} {sender} {name}>'

    @property
    def song_type(self):
        value = self._data.get('songInfo', {}).get('type')
        return value

    @property
    def song_external_id(self):
        value = self._data.get('songInfo', {}).get('fkid')
        return value

    @property
    def song_name(self):
        value = self._data.get('songInfo', {}).get('name')
        return value

    @property
    def song_id(self):
        value = self._data.get('songInfo', {}).get('songid')
        return value

    @property
    def length(self):
        value = self._data.get('songInfo', {}).get('songLength')
        if value:
            return datetime.timedelta(milliseconds=value)

    @property
    def played(self):
        value = self._data.get('song', {}).get('played')
        if value:
            return datetime.datetime.utcfromtimestamp(value / 1000)


class DubtrackJoin(DubtrackEvent):
    _data_type = 'user-join'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dubtrack_backend._register_user(self._data['user'])

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('user', {}).get('username'))

    def __repr__(self):
        cls = self.__class__.__name__
        sender = self.sender
        return f'<{cls} {sender}>'


class DubtrackUserPauseQueue(DubtrackEvent):
    _data_type = 'user-pause-queue'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dubtrack_backend._register_user(self._data['user'])

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('user', {}).get('username'))

    def __repr__(self):
        cls = self.__class__.__name__
        sender = self.sender
        return f'<{cls} {sender}>'


class DubtrackSetRole(DubtrackEvent):
    _data_type = 'user-setrole'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dubtrack_backend._register_user(self._data['user'])
        self._dubtrack_backend._register_user(self._data['modUser'])

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('user', {}).get('username'))

    @property
    def receiver(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('modUser', {}).get('_id'))

    def __repr__(self):
        cls = self.__class__.__name__
        role = self._data['role_object']['label']
        roletype = self._data['role_object']['type']
        rights = self._data['role_object']['rights']
        receiver = self.receiver
        sender = self.sender
        return f'<{cls} {receiver} -> {role}/{roletype}({", ".join(rights)}) by {sender}>'


class DubtrackUnSetRole(DubtrackEvent):
    _data_type = 'user-unsetrole'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dubtrack_backend._register_user(self._data['user'])
        self._dubtrack_backend._register_user(self._data['modUser'])

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('user', {}).get('username'))

    @property
    def receiver(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('modUser', {}).get('_id'))

    def __repr__(self):
        cls = self.__class__.__name__
        role = self._data['role_object']['label']
        roletype = self._data['role_object']['type']
        rights = self._data['role_object']['rights']
        receiver = self.receiver
        sender = self.sender
        return f'<{cls} {receiver} X {role}/{roletype}({", ".join(rights)}) by {sender}>'


class DubtrackUserUpdate(DubtrackEvent):
    _data_type = 'user_update'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dubtrack_backend._register_user(self._data['user'])

    @property
    def sender(self) -> DubtrackEntity:
        return self._dubtrack_backend._get_entity(self._data.get('user', {}).get('userid'))

    def __repr__(self):
        cls = self.__class__.__name__
        user = self._data['user']
        sender = self.sender
        skipped_count = user['skippedCount']
        played_count = user['playedCount']
        songs_in_queue = user['songsInQueue']
        dubs = user['dubs']
        return f'<{cls} {sender} skip#{skipped_count} played#{played_count} queue#{songs_in_queue} dubs#{dubs}>'


# Dubtrack bot plugin

class DubtrackBotBackend(Backend):
    # Official Bot methods
    def __init__(self):
        self.dubtrackws = DubtrackWS()
        self.dubtrack_channel = None
        self.dubtrack_users = defaultdict(dict)  # ID: user_session_info
        self.dubtrack_entities = weakref.WeakValueDictionary()
        self.dubtrack_id = None

    def configure(self, *, username=None, password=None):
        if any((username, password)):
            self.dubtrackws.set_login(username, password)
            ps = '*' * len(password)
            logger.debug(f'Setting username={username}, password={ps}')

    async def initialize(self):
        # Steps are
        await self.dubtrackws.initialize()
        if self.dubtrackws.logged_in:
            session_info = await self.dubtrackws.get_user_session_info()
            username = session_info['username']
            userid = session_info['userInfo']['userid']
            self.dubtrack_id = userid
            self._register_user(session_info)
            logger.info(f'Logged in as {username}#{userid}')
        else:
            logger.info(f'Connected, but not logged in')

        users = await self.dubtrackws.get_users()
        for user in users:
            self._register_user(user)

    async def consume(self):
        await self.dubtrackws.get_room_id()
        room_info = self.dubtrackws.room_info
        self.dubtrack_channel = DubtrackChannel(room_info, self)
        active_song = await self.dubtrackws.get_active_song()
        if active_song:
            yield DubtrackPlaying(active_song, self)
        async for data in self.dubtrackws.ws_api_consume():
            if data['type'].startswith('user_update'):
                event = DubtrackUserUpdate(data, self)
            else:
                event = DubtrackEvent.from_data(data, self)
            event.channel = self.dubtrack_channel
            yield event

    # Internal data tracking methods
    def _register_user(self, user_data):
        if not user_data:
            return
        user_id = None
        update_dict = {}
        if 'userInfo' in user_data:  # From user only
            if 'userid' in user_data['userInfo']:  # From profile
                user_id = user_data['userInfo']['userid']
            if 'created' in user_data:  # From profile only
                update_dict['created'] = user_data['created'] / 1000
        if 'userid' in user_data:  # From user-queue only
            user_id = user_data['userid']
            if 'dubs' in user_data:
                update_dict['dubs'] = user_data['dubs']
        if '_user' in user_data:
            if '_id' in user_data['_user']:
                user_id = user_data['_user']['_id']
            if 'username' in user_data['_user']:
                update_dict['username'] = user_data['_user']['username']
            if 'created' in user_data['_user']:  # From profile only
                update_dict['created'] = user_data['_user']['created'] / 1000
        if 'username' in user_data:  # From user/user-queue
            update_dict['username'] = user_data['username']
        if 'playedCount' in user_data:
            update_dict['played_count'] = user_data['playedCount']
        if 'songsInQueue' in user_data:
            update_dict['songs_in_queue'] = user_data['songsInQueue']
        if 'skippedCount' in user_data:
            update_dict['skips'] = user_data['skippedCount']

        if not user_id:
            return

        self.dubtrack_users[user_id].update(update_dict)

        # Update entity if it exists... As it seems the dict is not updated
        entity = self.dubtrack_entities.get(user_id)  # type: DubtrackEntity
        if entity:
            entity._data.update(update_dict)

    def _get_user_data(self, id_or_name):
        for id, user_data in self.dubtrack_users.items():
            if id_or_name == id:
                data = {'id': id}
                data.update(user_data)
                return data
            elif id_or_name == user_data.get('username'):
                data = {'id': id}
                data.update(user_data)
                return data

    def _get_entity(self, id_or_name):
        user_data = self._get_user_data(id_or_name)
        if not user_data:
            logger.info(f'Information for user {id_or_name} not available')
            return
        user_id = user_data['id']
        entity = self.dubtrack_entities.get(user_id)
        if entity:
            return entity
        entity = DubtrackEntity(user_data, self)
        self.dubtrack_entities[user_id] = entity
        return entity

    def whoami(self) -> DubtrackEntity:
        if self.dubtrack_id:
            return self._get_entity(self.dubtrack_id)


# Dubtrack dirty binding

def gen_request_id():
    return ''.join(random.choices(string.ascii_letters + string.digits, k=44))


class DubtrackWS:
    INIT = '0'
    PING = '2'
    PONG = '3'
    DATA = '4'

    def __init__(self, room='master-of-soundtrack'):
        self.room = room
        self.heartbeat = None
        self.ws_client_id = None
        self.ws_session = None
        self.connection_id = None
        self.connected_clients = defaultdict(set)
        self.aio_session = None
        self.user_session_info = None
        self.room_user_info = None
        self.room_info = None
        self.userpass = None
        self.suppress_messages = list()
        self.logged_in = None

    async def initialize(self):
        self.aio_session = aiohttp.ClientSession()
        # POST https://api.dubtrack.fm/auth/dubtrack
        if self.userpass:
            self.logged_in = await self.login(*self.userpass)
        # GET https://api.dubtrack.fm/auth/session
        await self.get_user_session_info()

    def set_login(self, username, password):
        if any((self.user_session_info, self.room_user_info, self.room_info)):
            raise ValueError('Once started, cannot login')
        self.userpass = (username, password)

    async def api_post(self, url, body):
        async with self.aio_session.post(url, body=body) as resp:
            logger.debug(f'Request: {url} - {body}')
            response = await resp.json()
            logger.debug(f'Response: {pprint.pformat(response)}')
        return response['data']

    async def api_get(self, url):
        async with self.aio_session.get(url) as resp:
            logger.debug(f'Request: {url}')
            response = await resp.json()
            logger.debug(f'Response: {pprint.pformat(response)}')
        return response['data']

    async def get_user_session_info(self):
        # {"userInfo": {"_id": "560b135c7ae1ea0300869b21",
        #               "userid": "560b135c7ae1ea0300869b20",
        #               "__v": 0},
        #  "_id": "560b135c7ae1ea0300869b20",
        #  "username": "txomon",
        #  "status": 1,
        #  "roleid": 1,
        #  "dubs": 0,
        #  "created": 1443566427591,
        #  "profileImage": {"public_id": "user/560b135c7ae1ea0300869b20",
        #                   "version": 1486657178,
        #                   "width": 245,
        #                   "height": 245,
        #                   "format": "gif",
        #                   "resource_type": "image",
        #                   "tags": [],
        #                   "pages": 22,
        #                   "bytes": 444903,
        #                   "type": "upload",
        #                   "etag": "09da0f0c34e6ddf6eb75516ea66e17bc",
        #                   "url": "http://res.cloudinary.com/hhberclba/image/upload/v1486657178/user"
        #                          "/560b135c7ae1ea0300869b20.gif",
        #                   "secure_url":
        #                       "https://res.cloudinary.com/hhberclba/image/upload/v1486657178/user"
        #                       "/560b135c7ae1ea0300869b20.gif",
        #                   "overwritten": True},
        #  "__v": 0}

        # OR

        # {'message': 'dologin'}

        if not self.user_session_info:
            self.user_session_info = await self.api_get('https://api.dubtrack.fm/auth/session')

        return self.user_session_info

    async def get_user_role(self):
        # {"room": {"_id": "561b1e59c90a9c0e00df610b",
        #           "name": "Master Of Soundtrack",
        #           "description": "film/videogame scores/soundtracks and epic/trailer music.\n\nrules: "
        #                          "http://mos.rf.gd/index.html\noverplayed list: "
        #                          "http://mos.rf.gd/overplayed.html\nfacebook page: "
        #                          "https://www.facebook.com/MasterofSoundtrack/\nmail address: masterofsoundtrack("
        #                          "@)yahoo.com\n---\ncustom skin: http://bit.ly/dubtracky\n---\ndub+, "
        #                          "helpful dubtrack extension: https://dub.plus/#/\n---\nnext NameThatTune: "
        #                          "https://www.facebook.com/events/203834513429820/\n---\nfirst ever MoS review: "
        #                          "http://bit.ly/296M8PX",
        #           "roomUrl": "master-of-soundtrack",
        #           "realTimeChannel": "dubtrackfm-master-of-soundtrack",
        #           "status": 1,
        #           "roomType": "room",
        #           "isPublic": False,
        #           "lang": None,
        #           "musicType": None,
        #           "password": None,
        #           "allowedDjs": 0,
        #           "maxLengthSong": 11,
        #           "displayQueue": False,
        #           "background": {"public_id": "kealmtphavj32zsshcsd",
        #                          "version": 1446588169,
        #                          "width": 1920,
        #                          "height": 1200,
        #                          "format": "jpg",
        #                          "resource_type": "image",
        #                          "tags": [],
        #                          "bytes": 849194,
        #                          "type": "upload",
        #                          "etag": "685925bd5248799f9f03eeeec7a485a1",
        #                          "url": "http://res.cloudinary.com/hhberclba/image/upload/v1446588169"
        #                                 "/kealmtphavj32zsshcsd.jpg",
        #                          "secure_url": "https://res.cloudinary.com/hhberclba/image/upload/v1446588169"
        #                                        "/kealmtphavj32zsshcsd.jpg"},
        #           "created": 1444617817050,
        #           "updated": 1444617817050,
        #           "userid": "56097db281c87803009bd1c5",
        #           "roomEmbed": "",
        #           "otSession": None,
        #           "_user": "56097db281c87803009bd1c5",
        #           "__v": 0,
        #           "activeUsers": 9,
        #           "currentSong": {"songid": "5677f5e03d9c59270031d943",
        #                           "type": "youtube",
        #                           "fkid": "XHaeAUFaHBU",
        #                           "name": "Halo Wars OST 01 Spirit of Fire"},
        #           "lockQueue": False,
        #           "welcomeMessage": "Welcome to MoS! Please take 5 minutes to read the rules: "
        #                             "http://mos.rf.gd/index.html - Also check the OP list and avoid those songs: "
        #                             "http://mos.rf.gd/overplayed.html\nVisit us on FB: http://xurl.es/hncih , "
        #                             "next NTT: http://xurl.es/3u71y",
        #           "metaDescription": "MoS, MasterOfSoundtrack, masterofsoundtrack, Master of Soundtrack, "
        #                              "master of soundtrack, scores, soundtrack, movie, film, videogame, "
        #                              "epic music, trailer",
        #           "displayInLobby": True,
        #           "displayInSearch": True,
        #           "limitQueueLength": False,
        #           "timeSongQueueRepeat": 1440,
        #           "recycleDJ": True,
        #           "displayDJinQueue": True,
        #           "displayUserJoin": True,
        #           "displayUserLeave": True,
        #           "displayUserGrab": True,
        #           "maxSongQueueLength": 0,
        #           "roomDisplay": "public",
        #           "waitListRandom": False,
        #           "slowMode": False},
        #  "user": {"_id": "5628db0a3883a45600b7e68f",
        #           "updated": 1519066844284,
        #           "skippedCount": 0,
        #           "playedCount": 1677,
        #           "songsInQueue": 0,
        #           "active": False,
        #           "dubs": 372,
        #           "order": 999999999,
        #           "roomid": "561b1e59c90a9c0e00df610b",
        #           "userid": "560b135c7ae1ea0300869b20",
        #           "_user": "560b135c7ae1ea0300869b20",
        #           "__v": 0,
        #           "ot_token": None,
        #           "authorized": True,
        #           "queuePaused": None,
        #           "waitLine": 0,
        #           "roleid": {"_id": "52d1ce33c38a06510c000001",
        #                      "type": "mod",
        #                      "label": "Moderator",
        #                      "rights": ["skip",
        #                                 "queue-order",
        #                                 "kick",
        #                                 "ban",
        #                                 "mute",
        #                                 "set-dj",
        #                                 "lock-queue",
        #                                 "delete-chat",
        #                                 "chat-mention"],
        #                      "__v": 0}}}

        if not self.room_user_info:
            room_id = await self.get_room_id()
            self.room_user_info = await self.api_post(f'https://api.dubtrack.fm/room/{room_id}/users', None)
        return self.room_user_info['user']['roleid']['type']

    async def say_in_room(self, text):
        # {"message": "pfff",
        #  "type": "chat-message",
        #  "chatid": None,
        #  "user": {"username": "txomon",
        #           "status": 1,
        #           "roleid": 1,
        #           "dubs": 0,
        #           "created": 1443566427591,
        #           "lastLogin": 0,
        #           "userInfo": {"_id": "560b135c7ae1ea0300869b21",
        #                        "userid": "560b135c7ae1ea0300869b20",
        #                        "__v": 0},
        #           "_force_updated": 1519046263690,
        #           "_id": "560b135c7ae1ea0300869b20",
        #           "__v": 0},
        #  "time": 1519047206095,
        #  "realTimeChannel": "dubtrackfm-master-of-soundtrack",
        #  "userRole": "mod"}

        body = {'chatid': None,
                'message': text,
                'time': int(time.time() * 1000),
                'type': 'chat-message',
                'user': self.user_session_info,
                'userRole': await self.get_user_role(), }
        room_id = await self.get_room_id()
        response = await self.aio_session.post(f'https://api.dubtrack.fm/chat/{room_id}', json=body)
        self.suppress_messages.append(text)
        return response

    async def login(self, username, password):
        # No response, just cookie set
        data = {'username': username, 'password': password}
        async with self.aio_session.post('https://api.dubtrack.fm/auth/dubtrack', data=data) as resp:
            return resp.status == 200

    async def get_token(self):
        # {'action': 17,
        #  'clientId': '4d3855621cd95d0d6a363806f712e2d0',
        #  'reqId': 'b5509b95b608cb7609891ef9e162cba3',
        #  'token': 'eyJhbGciOiJ.........HJwQL-7ytGnsXkucV5h_A_pV8V6YsA7DYQgxRpOaQMg'}
        response = await self.api_get('https://api.dubtrack.fm/auth/token')
        return response['token']

    async def get_room_id(self):
        # {'__v': 0,
        #  '_id': '561b1e59c90a9c0e00df610b',
        #  '_user': {'__v': 0,
        #            '_id': '56097db281c87803009bd1c5',
        #            'created': 1443462577850,
        #            'dubs': 0,
        #            'profileImage': {'bytes': 34965,
        #                             'etag': '495e80a989c26e9a5e672f243c8f3322',
        #                             'format': 'jpg',
        #                             'height': 500,
        #                             'overwritten': True,
        #                             'placeholder': False,
        #                             'resource_type': 'image',
        #                             'secure_url':
        # 'https://res.cloudinary.com/hhberclba/image/upload/v1510175888/user/56097db281c87803009bd1c5.jpg',
        #                             'type': 'upload',
        #                             'url':
        # 'http://res.cloudinary.com/hhberclba/image/upload/v1510175888/user/56097db281c87803009bd1c5.jpg',
        #                             'version': 1510175888,
        #                             'width': 500},
        #            'roleid': 1,
        #            'status': 1,
        #            'username': 'mullins'},
        #  'activeUsers': 9,
        #  'allowedDjs': 0,
        #  'background': {'bytes': 849194,
        #                 'etag': '685925bd5248799f9f03eeeec7a485a1',
        #                 'format': 'jpg',
        #                 'height': 1200,
        #                 'public_id': 'kealmtphavj32zsshcsd',
        #                 'resource_type': 'image',
        #                 'secure_url': 'https://res.cloudinary.com/hhberclba/image/upload/v1446588169
        # /kealmtphavj32zsshcsd.jpg',
        #                 'tags': [],
        #                 'type': 'upload',
        #                 'url': 'http://res.cloudinary.com/hhberclba/image/upload/v1446588169/kealmtphavj32zsshcsd
        # .jpg',
        #                 'version': 1446588169,
        #                 'width': 1920},
        #  'created': 1444617817050,
        #  'currentSong': {'fkid': '_imFm3FdFJE',
        #                  'name': 'Ennio Morricone - Giù la testa - (1971)',
        #                  'songid': '5824295e8fae8e0f00209a30',
        #                  'type': 'youtube'},
        #  'description': 'film/videogame scores/soundtracks and epic/trailer music.\n'
        #         '\n'
        #         'rules: http://mos.rf.gd/index.html\n'
        #         'overplayed list: http://mos.rf.gd/overplayed.html\n'
        #         'facebook page: https://www.facebook.com/MasterofSoundtrack/\n'
        #         'mail address: masterofsoundtrack(@)yahoo.com\n'
        #         '---\n'
        #         'custom skin: http://bit.ly/dubtracky\n'
        #         '---\n'
        #         'dub+, helpful dubtrack extension: https://dub.plus/#/\n'
        #         '---\n'
        #         'next NameThatTune: '
        #         'https://www.facebook.com/events/203834513429820/\n'
        #         '---\n'
        #         'first ever MoS review: http://bit.ly/296M8PX',
        #  'displayDJinQueue': True,
        #  'displayInLobby': True,
        #  'displayInSearch': True,
        #  'displayQueue': False,
        #  'displayUserGrab': True,
        #  'displayUserJoin': True,
        #  'displayUserLeave': True,
        #  'isPublic': False,
        #  'lang': None,
        #  'limitQueueLength': False,
        #  'lockQueue': False,
        #  'maxLengthSong': 11,
        #  'maxSongQueueLength': 0,
        #  'metaDescription': 'MoS, MasterOfSoundtrack, masterofsoundtrack, Master of '
        #  'Soundtrack, master of soundtrack, scores, soundtrack, '
        #  'movie, film, videogame, epic music, trailer',
        #  'musicType': None,
        #  'name': 'Master Of Soundtrack',
        #  'otSession': None,
        #  'password': None,
        #  'realTimeChannel': 'dubtrackfm-master-of-soundtrack',
        #  'recycleDJ': True,
        #  'roomDisplay': 'public',
        #  'roomEmbed': '',
        #  'roomType': 'room',
        #  'roomUrl': 'master-of-soundtrack',
        #  'slowMode': False,
        #  'status': 1,
        #  'timeSongQueueRepeat': 1440,
        #  'updated': 1444617817050,
        #  'userid': '56097db281c87803009bd1c5',
        #  'waitListRandom': False,
        #  'welcomeMessage': 'Welcome to MoS! Please take 5 minutes to read the rules: '
        #            'http://mos.rf.gd/index.html - Also check the OP list and '
        #            'avoid those songs: http://mos.rf.gd/overplayed.html\n'
        #            'Visit us on FB: http://xurl.es/hncih , next NTT: '
        #            'http://xurl.es/3u71y'}
        if not self.room_info:
            self.room_info = await self.api_get(f'https://api.dubtrack.fm/room/{self.room}')
        return self.room_info['_id']

    async def get_active_song(self):
        # {'song': {'__v': 0,
        #           '_id': '5a8be8a5edab940100b2af20',
        #           '_song': '583198efc2225d2e00849a5b',
        #           '_user': '562e462b888e6f1900ff20bc',
        #           'created': 1519118499376,
        #           'downdubs': 0,
        #           'isActive': True,
        #           'isPlayed': False,
        #           'order': 1,
        #           'played': 1519131687972,
        #           'roomid': '561b1e59c90a9c0e00df610b',
        #           'skipped': False,
        #           'songLength': 307000,
        #           'songid': '583198efc2225d2e00849a5b',
        #           'updubs': 2,
        #           'userid': '562e462b888e6f1900ff20bc'},
        #  'songInfo': {'__v': 0,
        #               '_id': '583198efc2225d2e00849a5b',
        #               'created': '2016-11-20T12:37:03.018Z',
        #               'fkid': 'zoORlnu_6dk',
        #               'images': {'thumbnail': 'https://i.ytimg.com/vi/zoORlnu_6dk/hqdefault.jpg'},
        #               'name': 'Spain Ambient - Guitar Improv 7 (Civilization 6 OST)',
        #               'songLength': 307000,
        #               'type': 'youtube'},
        #  'startTime': 65}

        # or if no one is playing

        # {'err': {'details': {'code': 404,
        #                      'message': 'no songs in active queue'},
        #          'origin': 'method'}}

        room_id = await self.get_room_id()
        playing_song = await self.api_get(f'https://api.dubtrack.fm/room/{room_id}/playlist/active')
        if 'err' in playing_song:
            return None
        return playing_song

    async def get_users(self):
        # [
        #     {"_id": "57f36aff34169c1a0018f92d",
        #      "updated": 1519126214587,
        #      "skippedCount": 0,
        #      "playedCount": 10894,
        #      "songsInQueue": 556,
        #      "active": True,
        #      "dubs": 3542,
        #      "order": 99999,
        #      "roomid": "561b1e59c90a9c0e00df610b",
        #      "userid": "57f36acd6c9b5c5b003d41d2",
        #      "_user": {
        #          "_id": "57f36acd6c9b5c5b003d41d2",
        #          "username": "eberg",
        #          "status": 1,
        #          "roleid": 1,
        #          "dubs": 0,
        #          "created": 1475570381585,
        #          "profileImage": {
        #              "public_id": "user/57f36acd6c9b5c5b003d41d2",
        #              "version": 1476084566, "width": 500, "height": 267, "format": "jpg", "resource_type": "image",
        #              "tags": [], "bytes": 47184, "type": "upload", "etag": "010154daa60a012e27cfc5ef888dae00",
        #              "url": "http://res.cloudinary.com/hhberclba/image/upload/v1476084566/user"
        #                     "/57f36acd6c9b5c5b003d41d2.jpg",
        #              "secure_url": "https://res.cloudinary.com/hhberclba/image/upload/v1476084566/user"
        #                            "/57f36acd6c9b5c5b003d41d2.jpg",
        #              "overwritten": True},
        #          "__v": 0},
        #      "__v": 0,
        #      "authorized": True,
        #      "ot_token": None,
        #      "queuePaused": None,
        #      "waitLine": 0},
        # ]
        room_id = await self.get_room_id()
        return await self.api_get(f'https://api.dubtrack.fm/room/{room_id}/users')

    async def get_user(self, user_id):
        # {'__v': 0,
        #  '_id': '5628edc08d7d6a5600335d3d',
        #  'created': 1445522880666,
        #  'dubs': 0,
        #  'profileImage': {
        #      'bytes': 61889,
        #      'etag': 'f5f398bf49c71314e865ade3c4a7bf2a',
        #      'format': 'jpg',
        #      'height': 455,
        #      'overwritten': True,
        #      'placeholder': False,
        #      'public_id': 'user/5628edc08d7d6a5600335d3d',
        #      'resource_type': 'image',
        #      'secure_url': 'https://res.cloudinary.com/hhberclba/image/upload/v1508887786/user'
        #                    '/5628edc08d7d6a5600335d3d.jpg',
        #      'tags': [],
        #      'type': 'upload',
        #      'url': 'http://res.cloudinary.com/hhberclba/image/upload/v1508887786/user'
        #             '/5628edc08d7d6a5600335d3d.jpg',
        #      'version': 1508887786,
        #      'width': 455
        #  },
        #  'roleid': 1,
        #  'status': 1,
        #  'userInfo': {
        #      '__v': 0,
        #      '_id': '5628edc08d7d6a5600335d3e',
        #      'userid': '5628edc08d7d6a5600335d3d'
        #  },
        #  'username': 'iCel'
        #  }
        return await self.api_get(f'https://api.dubtrack.fm/user/{user_id}')

    async def get_history(self, page=None):
        # [{'__v': 0,
        #   '_id': '5a8453aed51c3101003def50',
        #   '_song': {'__v': 0,
        #             '_id': '5795e51c828c22790037db1e',
        #             'created': '2016-07-25T10:08:28.691Z',
        #             'fkid': 'EqkFgAn4U-o',
        #             'images': {'thumbnail': 'https://i.ytimg.com/vi/EqkFgAn4U-o/hqdefault.jpg'},
        #             'name': 'Me Before You Orchestral- Craig Armstrong (Me Before You- The Score)',
        #             'songLength': 434000,
        #             'type': 'youtube'},
        #   '_user': {'__v': 0,
        #             '_id': '56a80c626894b9410067b716',
        #             'created': 1453853793914,
        #             'dubs': 0,
        #             'profileImage': {'bytes': 33387,
        #                              'etag': '0eb4420cbced52aa81f1bd3368a87f27',
        #                              'format': 'gif',
        #                              'height': 325,
        #                              'pages': 4,
        #                              'public_id': 'user/56a80c626894b9410067b716',
        #                              'resource_type': 'image',
        #                              'secure_url':
        # 'https://res.cloudinary.com/hhberclba/image/upload/v1477307984/user/56a80c626894b9410067b716.gif',
        #                              'tags': [],
        #                              'type': 'upload',
        #                              'url':
        # 'http://res.cloudinary.com/hhberclba/image/upload/v1477307984/user/56a80c626894b9410067b716.gif',
        #                              'version': 1477307984,
        #                              'width': 325},
        #             'roleid': 1,
        #             'status': 1,
        #             'username': 'hennersC'},
        #     'created': 1518621605692,
        #     'downdubs': 0,
        #     'isActive': True,
        #     'isPlayed': True,
        #     'order': 10,
        #     'played': 1518787795607,
        #     'roomid': '561b1e59c90a9c0e00df610b',
        #     'skipped': False,
        #     'songLength': 434000,
        #     'songid': '5795e51c828c22790037db1e',
        #     'updubs': 2,
        #     'userid': '56a80c626894b9410067b716'}]
        room_id = await self.get_room_id()
        url = URL(f'https://api.dubtrack.fm/room/{room_id}/playlist/history')
        if page:
            url = url.with_query({'page': page})
        return await self.api_get(str(url))

    async def get_room_playlist(self):
        # {"__v": 0,
        #  "created": 1520158007231,
        #  "isActive": False,
        #  "isPlayed": False,
        #  "skipped": False,
        #  "order": 999,
        #  "roomid": "561b1e59c90a9c0e00df610b",
        #  "songLength": 168000,
        #  "updubs": 0,
        #  "downdubs": 0,
        #  "userid": "560b135c7ae1ea0300869b20",
        #  "songid": "565e8fb8e39987830055f389",
        #  "_user": "560b135c7ae1ea0300869b20",
        #  "_song": "565e8fb8e39987830055f389",
        #  "_id": "5a9bc537b55af20100405d6b"}
        room_id = await self.get_room_id()
        playlist = await self.api_get(f'https://api.dubtrack.fm/room/{room_id}/playlist')
        return playlist

    async def add_song_to_playlist(self, extid, origin='youtube'):
        # {"__v": 0,
        #  "created": 1520158645688,
        #  "isActive": False,
        #  "isPlayed": False,
        #  "skipped": False,
        #  "order": 999,
        #  "roomid": "561b1e59c90a9c0e00df610b",
        #  "songLength": 222000,
        #  "updubs": 0,
        #  "downdubs": 0,
        #  "userid": "560b135c7ae1ea0300869b20",
        #  "songid": "58b6a2126ba2fa18005f3a8e",
        #  "_user": "560b135c7ae1ea0300869b20",
        #  "_song": "58b6a2126ba2fa18005f3a8e",
        #  "_id": "5a9bc7b58befc60100230823"}
        room_id = await self.get_room_id()
        response = self.api_post(f'https://api.dubtrack.fm/room/{room_id}/playlist',
                                 {'songId': extid, 'songType': origin})
        return response

    async def get_room_playlist_details(self):
        # [
        #     {"_id": "5a9bc7ba9d993401003eba43",
        #      "created": 1520158650527,
        #      "isActive": False,
        #      "isPlayed": False,
        #      "skipped": False,
        #      "order": 999,
        #      "roomid": "561b1e59c90a9c0e00df610b",
        #      "songLength": 309000,
        #      "updubs": 0,
        #      "downdubs": 0,
        #      "userid": "560b135c7ae1ea0300869b20",
        #      "songid": "5a9bc7ba9d993401003eba42",
        #      "_user": {"_id": "560b135c7ae1ea0300869b20",
        #                "username": "txomon",
        #                "status": 1,
        #                "roleid": 1,
        #                "dubs": 0,
        #                "created": 1443566427591,
        #                "profileImage": {"public_id": "user/560b135c7ae1ea0300869b20",
        #                                 "version": 1486657178,
        #                                 "width": 245,
        #                                 "height": 245,
        #                                 "format": "gif",
        #                                 "resource_type": "image",
        #                                 "tags": [],
        #                                 "pages": 22,
        #                                 "bytes": 444903,
        #                                 "type": "upload",
        #                                 "etag": "09da0f0c34e6ddf6eb75516ea66e17bc",
        #                                 "url":
        #                                     "http://res.cloudinary.com/hhberclba/image/upload/v1486657178/user"
        #                                     "/560b135c7ae1ea0300869b20.gif",
        #                                 "secure_url":
        #                                     "https://res.cloudinary.com/hhberclba/image/upload/v1486657178/user"
        #                                     "/560b135c7ae1ea0300869b20.gif",
        #                                 "overwritten": True},
        #                "__v": 0},
        #      "_song": {"_id": "5a9bc7ba9d993401003eba42",
        #                "name": "Chu Ishikawa - Megatron (Clean Version)",
        #                "images": {"thumbnail": "https://i.ytimg.com/vi/Nku49pPj0kU/hqdefault.jpg"},
        #                "type": "youtube",
        #                "songLength": 309000,
        #                "fkid": "Nku49pPj0kU",
        #                "__v": 0,
        #                "created": "2018-03-04T10:17:30.329Z"},
        #      "__v": 0}]
        room_id = await self.get_room_id()
        response = self.api_get(f'https://api.dubtrack.fm/room/{room_id}/playlist/details')
        return response

    async def delete_track_in_queue(self, user_id):
        # {"userNextSong": {
        #     "_id": "5a9bce0357faec01003e0c03",
        #     "created": 1520160259582,
        #     "isActive": False,
        #     "isPlayed": False,
        #     "skipped": False,
        #     "order": 999,
        #     "roomid": "561b1e59c90a9c0e00df610b",
        #     "songLength": 246000,
        #     "updubs": 0,
        #     "downdubs": 0,
        #     "userid": "560b135c7ae1ea0300869b20",
        #     "songid": "560895658e9cb60300550def",
        #     "_user": {
        #         "_id": "560b135c7ae1ea0300869b20",
        #         "username": "txomon",
        #         "status": 1,
        #         "roleid": 1,
        #         "dubs": 0,
        #         "created": 1443566427591,
        #         "profileImage": {
        #             "public_id": "user/560b135c7ae1ea0300869b20",
        #             "version": 1486657178,
        #             "width": 245,
        #             "height": 245,
        #             "format": "gif",
        #             "resource_type": "image",
        #             "tags": [],
        #             "pages": 22,
        #             "bytes": 444903,
        #             "type": "upload",
        #             "etag": "09da0f0c34e6ddf6eb75516ea66e17bc",
        #             "url":
        #                 "http://res.cloudinary.com/hhberclba/image/upload/v1486657178/user"
        #                 "/560b135c7ae1ea0300869b20.gif",
        #             "secure_url":
        #                 "https://res.cloudinary.com/hhberclba/image/upload/v1486657178/user"
        #                 "/560b135c7ae1ea0300869b20.gif",
        #             "overwritten": True},
        #         "__v": 0},
        #     "_song": {
        #         "_id": "560895658e9cb60300550def",
        #         "name": "Back To The Future - The Power Of Love",
        #         "images": {
        #             "thumbnail": "https://i.ytimg.com/vi/-NMph943tsw/hqdefault.jpg",
        #             "youtube": {"default": {
        #                 "url": "https://i.ytimg.com/vi/-NMph943tsw/default.jpg",
        #                 "width": 120,
        #                 "height": 90},
        #                 "medium": {
        #                     "url": "https://i.ytimg.com/vi/-NMph943tsw/mqdefault.jpg",
        #                     "width": 320,
        #                     "height": 180},
        #                 "high": {
        #                     "url": "https://i.ytimg.com/vi/-NMph943tsw/hqdefault.jpg",
        #                     "width": 480,
        #                     "height": 360},
        #                 "standard": {
        #                     "url": "https://i.ytimg.com/vi/-NMph943tsw/sddefault.jpg",
        #                     "width": 640,
        #                     "height": 480}}},
        #         "type": "youtube",
        #         "fkid": "-NMph943tsw",
        #         "streamUrl": None,
        #         "songLength": 246000,
        #         "__v": 0,
        #         "created": "2015-09-28T01:18:29.979Z"},
        #     "__v": 0}}

        # or if no next song in user's queue

        # {"userNextSong": None}

        room_id = await self.get_room_id()
        url = f'https://api.dubtrack.fm/room/{room_id}/queue/user/{user_id}'
        async with self.aio_session.delete(url) as resp:
            logger.debug(f'Request: {url}')
            response = await resp.json()
            logger.debug(f'Response: {pprint.pformat(response)}')
        return response['data']

    async def raw_ws_consume(self):
        last_token_fail = last_consume_fail = 0
        while True:  # two tries per level
            try:
                token = await self.get_token()
            except Exception:
                logger_layer1.exception('Trouble getting token')
                now = time.time()
                if last_token_fail + 15 > now:  # If the token has failed in the last 15 secs
                    break
                last_token_fail, last_consume_fail = now, 0
                continue
            while True:
                try:
                    async for msg in self._raw_ws_consume(access_token=token):
                        yield msg
                except Exception:
                    logger_layer1.exception('Consumption has failed')
                    now = time.time()
                    if last_consume_fail + 15 > now:  # If consumption has failed in the last 15 secs
                        break
                    last_consume_fail = now

    async def _raw_ws_consume(self, access_token):
        params = {
            'connect': 1,
            'EIO': 3,
            'transport': 'websocket',
            'access_token': access_token,
        }
        if self.ws_client_id:
            params['clientId'] = self.ws_client_id

        ws_connect_url = str(URL('https://ws.dubtrack.fm/ws/').with_query(params))

        async with self.aio_session.ws_connect(ws_connect_url) as ws_session:
            self.ws_session = ws_session
            await self.ws_session_opened_cb()
            async for msg in ws_session:
                if msg.type in (aiohttp.WSMsgType.CLOSING, aiohttp.WSMsgType.CLOSED):
                    logger_layer1.debug('Closed WS channel')
                    break
                elif msg.type == aiohttp.WSMsgType.TEXT:
                    yield (ws_session, msg.data)
            self.ws_session = None

    async def ws_send(self, message):
        for _ in range(3):
            if self.ws_session:
                break
            await asyncio.sleep(0.5)
        else:
            logger_layer1.error('Cannot send message, ws_session is not set')
            logger_layer1.debug(f'Message that cannot be sent is: {message}')
            raise Exception('No session available')
        logger_layer1.debug(f'Sending message {message}')
        await self.ws_session.send_str(message)

    async def do_heartbeat(self, interval):
        while True:
            logger_layer1.debug('Sending ping')
            await self.ws_send(self.PING)
            await asyncio.sleep(interval / 1000)

    async def send_room_subscription(self):
        room_id = await self.get_room_id()
        subscription = {
            'action': 10,
            "channel": f'room:{room_id}',
        }
        await self.ws_send(f'4{json.dumps(subscription)}')

    async def send_presence_update(self):
        room_id = await self.get_room_id()
        presence_update = {
            "action": 14,
            "channel": f'room:{room_id}',
            "presence": {"action": 0, "data": {}},
            "reqId": gen_request_id()}
        await self.ws_send(f'4{json.dumps(presence_update)}')

    async def ws_session_opened_cb(self):
        if not self.heartbeat:
            # Just harcoding, not important
            self.heartbeat = asyncio.ensure_future(self.do_heartbeat(25000))
        await self.send_room_subscription()
        await self.send_presence_update()

    async def ws_api_consume(self):
        async for session, message in self.raw_ws_consume():
            # First layer
            # They have a digit+json... => 1{"asdf": "czxc"}
            code = message[0]
            if code == self.INIT:
                continue  # Ignoring the heartbeat rate for now
            elif code == self.PING:
                logger_layer1.warning('Received a ping?!?')
                continue
            elif code == self.PONG:
                logger_layer1.debug('Received pong')
                continue
            elif code != '4':  # 4 is the main case
                logger_layer1.warning('Received unknown message {message}')
                continue

            logger_layer1.debug(f'Received message: {message}')

            # Second layer: 4
            # They have action in the first level of the json
            data = json.loads(message[1:])
            action = data['action']
            if action == 4:  # Client ID given for future reconnections
                self.ws_client_id = data['clientId']
                self.connection_id = data['connectionId']
                continue
            elif action == 11:  # ACK from action=10, also contains what in #4
                continue
            elif action == 14:  # Presence updates
                presence = data['presence']
                connection_id = presence.get('connectionId')
                client_id = presence.get('clientId')
                if 'reqId' in data and connection_id != self.connection_id:
                    logger_layer2.error(
                        f'Presence packet says connectionId {connection_id} instead of {self.connection_id}. '
                        f'Ignoring..?')
                    continue
                if action == 0:
                    logger_layer2.debug(
                        f'Client {client_id} connected with {connection_id}')
                    self.connected_clients[client_id].add(connection_id)
                elif action == 1:
                    logger_layer2.debug(
                        f'Client {client_id} disconnected with {connection_id}')
                    self.connected_clients[client_id].remove(connection_id)
                continue
            elif action != 15:  # 4, Action 15 is the main case
                logger_layer2.warning('Received unknown action {action}')
                continue

            # Third layer: 4=>action#15
            message = data['message']
            if message['type'] != 'json':
                logger_layer3.info(
                    f'Ignoring, becase type is not json: {message}')
                continue

            content_type = message['name']
            content = json.loads(message['data'])
            if content_type == 'chat-message':
                # {'chatid': '560b135c7ae1ea0300869b20-1518783003490',
                #  'message': 'this is going goood :P',
                #  'queue_object': {'__v': 0,
                #                   '_id': '5628db0a3883a45600b7e68f',
                #                   '_user': '560b135c7ae1ea0300869b20',
                #                   'active': True,
                #                   'authorized': True,
                #                   'dubs': 368,
                #                   'order': 99999,
                #                   'ot_token': None,
                #                   'playedCount': 1677,
                #                   'queuePaused': None,
                #                   'roleid': '52d1ce33c38a06510c000001',
                #                   'roomid': '561b1e59c90a9c0e00df610b',
                #                   'skippedCount': 0,
                #                   'songsInQueue': 0,
                #                   'updated': 1518771989676,
                #                   'userid': '560b135c7ae1ea0300869b20',
                #                   'waitLine': 0},
                #  'time': 1518783003490,
                #  'type': 'chat-message',
                #  'user': {'__v': 0,
                #           '_force_updated': 1516971162191,
                #           '_id': '560b135c7ae1ea0300869b20',
                #           'created': 1443566427591,
                #           'dubs': 0,
                #           'roleid': 1,
                #           'status': 1,
                #           'userInfo': {'__v': 0,
                #                        '_id': '560b135c7ae1ea0300869b21',
                #                        'userid': '560b135c7ae1ea0300869b20'},
                #           'username': 'txomon'}}
                chatid = content['chatid']
                username = content['user']['username']
                userid = content['user']['userInfo']['userid']
                msg = content["message"]
                logger_layer3.debug(
                    f'Chat {username}#{userid} (chatid#{chatid}): {msg}')
                if msg in self.suppress_messages:
                    self.suppress_messages.remove(msg)
                    logger_layer3.debug(f'Suppressing message: {msg}')
                    continue
            elif content_type == 'chat-skip':
                # {'type': 'chat-skip', 'username': 'txomon'}
                username = content['username']
                logger_layer3.debug(f'Chat-skip by {username}')
            elif content_type == 'delete-chat-message':
                # {'chatid': '560b135c7ae1ea0300869b20-1518784684020',
                #  'type': 'delete-chat-message',
                #  'user': {'__v': 0,
                #           '_force_updated': 1516971162191,
                #           '_id': '560b135c7ae1ea0300869b20',
                #           'created': 1443566427591,
                #           'dubs': 0,
                #           'roleid': 1,
                #           'status': 1,
                #           'userInfo': {'__v': 0,
                #                        '_id': '560b135c7ae1ea0300869b21',
                #                        'userid': '560b135c7ae1ea0300869b20'},
                #           'username': 'txomon'}}
                username = content['user']['username']
                userid = content['user']['userInfo']['userid']
                chatid = content['chatid']
                logger_layer3.debug(f'User {username}#{userid} deleted {chatid}')
            elif content_type == 'room_playlist-dub':
                # {'dubtype': 'downdub',
                #  'playlist': {'__v': 0,
                #               '_id': '5a8453cad51c3101003df01c',
                #               '_song': '5a0db972fd20620100678621',
                #               '_user': '56a80c626894b9410067b716',
                #               'created': 1518621640481,
                #               'downdubs': 1,
                #               'isActive': True,
                #               'isPlayed': False,
                #               'order': 2,
                #               'played': 1518782587986,
                #               'roomid': '561b1e59c90a9c0e00df610b',
                #               'skipped': False,
                #               'songLength': 221000,
                #               'songid': '5a0db972fd20620100678621',
                #               'updubs': 0,
                #               'userid': '56a80c626894b9410067b716'},
                #  'type': 'room_playlist-dub',
                #  'user': {'__v': 0,
                #           '_force_updated': 1516971162191,
                #           '_id': '560b135c7ae1ea0300869b20',
                #           'created': 1443566427591,
                #           'dubs': 0,
                #           'roleid': 1,
                #           'status': 1,
                #           'userInfo': {'__v': 0,
                #                        '_id': '560b135c7ae1ea0300869b21',
                #                        'userid': '560b135c7ae1ea0300869b20'},
                #           'username': 'txomon'}}
                dubtype = content['dubtype']
                username = content['user']['username']
                userid = content['user']['userInfo']['userid']
                downdubs = content['playlist']['downdubs']
                updubs = content['playlist']['updubs']
                logger_layer3.debug(
                    f"Song {dubtype} by {username}#{userid}, total {updubs}U/{downdubs}D")
            elif content_type == 'room_playlist-queue-reorder':
                # {'type': 'room_playlist-queue-reorder',
                #  'user': {'__v': 0,
                #           '_force_updated': 1516971162191,
                #           '_id': '560b135c7ae1ea0300869b20',
                #           'created': 1443566427591,
                #           'dubs': 0,
                #           'roleid': 1,
                #           'status': 1,
                #           'userInfo': {'__v': 0,
                #                        '_id': '560b135c7ae1ea0300869b21',
                #                        'userid': '560b135c7ae1ea0300869b20'},
                #           'username': 'txomon'}}
                username = content['user']['username']
                userid = content['user']['userInfo']['userid']
                logger_layer3.debug(
                    f'User {username}#{userid} reordered the queue')
            elif content_type == 'room_playlist-queue-update-dub':
                # {'type': 'room_playlist-queue-update-dub',
                #  'user': {'__v': 0,
                #           '_force_updated': 1499443841892,
                #           '_id': '5628edc08d7d6a5600335d3d',
                #           'created': 1445522880666,
                #           'dubs': 0,
                #           'roleid': 1,
                #           'status': 1,
                #           'userInfo': {'__v': 0,
                #                        '_id': '5628edc08d7d6a5600335d3e',
                #                        'userid': '5628edc08d7d6a5600335d3d'},
                #           'username': 'iCel'}}
                username = content['user']['username']
                userid = content['user']['userInfo']['userid']
                logger_layer3.debug(
                    f'User {username}/{userid} changed personal queue')
            elif content_type == 'room_playlist-update':
                # {'startTime': -1,
                #  'song': {'_id': '5a853a0a07f061010053d3c8',
                #           'created': 1518680579959,
                #           'isActive': True,
                #           'isPlayed': False,
                #           'skipped': False,
                #           'order': 109,
                #           'roomid': '561b1e59c90a9c0e00df610b',
                #           'songLength': 194000,
                #           'updubs': 0, 'downdubs': 0,
                #           'userid': '56096ce7a98a6b0300144e33',
                #           'songid': '584efe5534194d8400cfd013',
                #           '_user': '56096ce7a98a6b0300144e33',
                #           '_song': '584efe5534194d8400cfd013',
                #           '__v': 0,
                #           'played': 1518781826209},
                #  'songInfo': {'_id': '584efe5534194d8400cfd013',
                #               'name': 'Luis Alvarez - Final Time',
                #               'images': {'thumbnail': 'https://i.ytimg.com/vi/KHkPbbdu5kk/hqdefault.jpg'},
                #               'type': 'youtube',
                #               'songLength': 194000,
                #               'fkid': 'KHkPbbdu5kk',
                #               '__v': 0,
                #               'created': '2016-12-12T19:45:25.669Z'},
                #  'type': 'room_playlist-update'}
                songinfo = content['songInfo']
                name = songinfo['name']
                songtype = songinfo['type']
                songid = songinfo['fkid']
                logger_layer3.debug(f'Now playing {songtype}#{songid}: {name}')
            elif content_type == 'user-join':
                # {'roomUser': {'__v': 0,
                #               '_id': '57f36aff34169c1a0018f92d',
                #               '_user': '57f36acd6c9b5c5b003d41d2',
                #               'active': False,
                #               'authorized': True,
                #               'dubs': 3533,
                #               'order': 99999,
                #               'ot_token': None,
                #               'playedCount': 10862,
                #               'queuePaused': None,
                #               'roomid': '561b1e59c90a9c0e00df610b',
                #               'skippedCount': 0,
                #               'songsInQueue': 588,
                #               'updated': 1518783589638,
                #               'userid': '57f36acd6c9b5c5b003d41d2',
                #               'waitLine': 0},
                #  'type': 'user-join',
                #  'user': {'__v': 0,
                #           '_force_updated': 1504509671098,
                #           '_id': '57f36acd6c9b5c5b003d41d2',
                #           'created': 1475570381585,
                #           'dubs': 0,
                #           'roleid': 1,
                #           'status': 1,
                #           'userInfo': {'__v': 0,
                #                        '_id': '57f36acd6c9b5c5b003d41d3',
                #                        'userid': '57f36acd6c9b5c5b003d41d2'},
                #           'username': 'eberg'}}
                username = content['user']['username']
                userid = content['user']['userInfo']['userid']
                # TODO: Explore roomUser
                logger_layer3.debug(f'User {username}#{userid} joined')
            elif content_type == 'user-pause-queue':
                # {'type': 'user-pause-queue',
                #  'user': {'__v': 0,
                #           '_force_updated': 1499443841892,
                #           '_id': '5628edc08d7d6a5600335d3d',
                #           'created': 1445522880666,
                #           'dubs': 0,
                #           'roleid': 1,
                #           'status': 1,
                #           'userInfo': {'__v': 0,
                #                        '_id': '5628edc08d7d6a5600335d3e',
                #                        'userid': '5628edc08d7d6a5600335d3d'},
                #           'username': 'iCel'},
                # 'user_queue': {'__v': 0,
                #                '_id': '5628ededa2d0f81300edc39a',
                #                '_user': '5628edc08d7d6a5600335d3d',
                #                'active': True,
                #                'authorized': True,
                #                'dubs': 25899,
                #                'order': 99999,
                #                'ot_token': None,
                #                'playedCount': 23897,
                #                'queuePaused': None,
                #                'roleid': '5615fa9ae596154a5c000000',
                #                'roomid': '561b1e59c90a9c0e00df610b',
                #                'skippedCount': 0,
                #                'songsInQueue': 31,
                #                'updated': 1518783663567,
                #                'userid': '5628edc08d7d6a5600335d3d',
                #                'waitLine': 0}}

                # OR

                # {'type': 'user-pause-queue',
                #  'user': {'__v': 0,
                #           '_id': '560b135c7ae1ea0300869b20',
                #           'created': 1443566427591,
                #           'dubs': 0,
                #           'profileImage': {'bytes': 444903,
                #                            'etag': '09da0f0c34e6ddf6eb75516ea66e17bc',
                #                            'format': 'gif',
                #                            'height': 245,
                #                            'overwritten': True,
                #                            'pages': 22,
                #                            'public_id': 'user/560b135c7ae1ea0300869b20',
                #                            'resource_type': 'image',
                #                            'secure_url':
                # 'https://res.cloudinary.com/hhberclba/image/upload/v1486657178/user/560b135c7ae1ea0300869b20.gif',
                #                            'tags': [],
                #                            'type': 'upload',
                #                            'url':
                # 'http://res.cloudinary.com/hhberclba/image/upload/v1486657178/user/560b135c7ae1ea0300869b20.gif',
                #                            'version': 1486657178,
                #                            'width': 245},
                #           'roleid': 1,
                #           'status': 1,
                #           'username': 'txomon'},
                #  'user_queue': {'__v': 0,
                #                 '_id': '5628db0a3883a45600b7e68f',
                #                 '_user': '560b135c7ae1ea0300869b20',
                #                 'active': True,
                #                 'authorized': True,
                #                 'dubs': 376,
                #                 'order': 99999,
                #                 'ot_token': None,
                #                 'playedCount': 1680,
                #                 'queuePaused': None,
                #                 'roleid': '52d1ce33c38a06510c000001',
                #                 'roomid': '561b1e59c90a9c0e00df610b',
                #                 'skippedCount': 0,
                #                 'songsInQueue': 0,
                #                 'updated': 1519080836290,
                #                 'userid': '560b135c7ae1ea0300869b20',
                #                 'waitLine': 0}}

                # TODO: Correct for both posibilities
                username = content['user']['username']
                userid = content['user'].get('userInfo', {}).get('userid')
                userid = userid or content['user']['_id']

                logger_layer3.debug(f'User {username}#{userid} stopped playlist')
            elif content_type == 'user-setrole':
                # {'modUser': {'__v': 0,
                #              '_id': '560b135c7ae1ea0300869b20',
                #              'created': 1443566427591,
                #              'dubs': 0,
                #              'profileImage': {'bytes': 444903,
                #                               'etag': '09da0f0c34e6ddf6eb75516ea66e17bc',
                #                               'format': 'gif',
                #                               'height': 245,
                #                               'overwritten': True,
                #                               'pages': 22,
                #                               'public_id': 'user/560b135c7ae1ea0300869b20',
                #                               'resource_type': 'image',
                #                               'secure_url':
                # 'https://res.cloudinary.com/hhberclba/image/upload/v1486657178/user/560b135c7ae1ea0300869b20.gif',
                #                               'tags': [],
                #                               'type': 'upload',
                #                               'url':
                # 'http://res.cloudinary.com/hhberclba/image/upload/v1486657178/user/560b135c7ae1ea0300869b20.gif',
                #                               'version': 1486657178,
                #                               'width': 245},
                #              'roleid': 1,
                #              'status': 1,
                #              'username': 'txomon'},
                #  'role_object': {'__v': 0,
                #                  '_id': '52d1ce33c38a06510c000001',
                #                  'label': 'Moderator',
                #                  'rights': ['skip',
                #                             'queue-order',
                #                             'kick',
                #                             'ban',
                #                             'mute',
                #                             'set-dj',
                #                             'lock-queue',
                #                             'delete-chat',
                #                             'chat-mention'],
                #                  'type': 'mod'},
                #  'type': 'user-setrole',
                #  'user': {'__v': 0,
                #           '_force_updated': 1499443841892,
                #           '_id': '5628edc08d7d6a5600335d3d',
                #           'created': 1445522880666,
                #           'dubs': 0,
                #           'roleid': 1,
                #           'status': 1,
                #           'userInfo': {'__v': 0,
                #                        '_id': '5628edc08d7d6a5600335d3e',
                #                        'userid': '5628edc08d7d6a5600335d3d'},
                #           'username': 'iCel'}}
                modname = content['modUser']['username']
                modid = content['modUser']['_id']
                role = content['role_object']['label']
                roletype = content['role_object']['type']
                rights = content['role_object']['rights']
                username = content['user']['username']
                userid = content['user']['userInfo']['userid']
                logger_layer3.debug(
                    f'User {modname}#{modid} moved to role {role}/{roletype}({", ".join(rights)}) by {username}#'
                    f'{userid}')
            elif content_type == 'user-unsetrole':
                # {'modUser': {'__v': 0,
                #              '_id': '560b135c7ae1ea0300869b20',
                #              'created': 1443566427591,
                #              'dubs': 0,
                #              'profileImage': {'bytes': 444903,
                #                               'etag': '09da0f0c34e6ddf6eb75516ea66e17bc',
                #                               'format': 'gif',
                #                               'height': 245,
                #                               'overwritten': True,
                #                               'pages': 22,
                #                               'public_id': 'user/560b135c7ae1ea0300869b20',
                #                               'resource_type': 'image',
                #                               'secure_url':
                # 'https://res.cloudinary.com/hhberclba/image/upload/v1486657178/user/560b135c7ae1ea0300869b20.gif',
                #                               'tags': [],
                #                               'type': 'upload',
                #                               'url':
                # 'http://res.cloudinary.com/hhberclba/image/upload/v1486657178/user/560b135c7ae1ea0300869b20.gif',
                #                               'version': 1486657178,
                #                               'width': 245},
                #              'roleid': 1,
                #              'status': 1,
                #              'username': 'txomon'},
                #  'role_object': {'__v': 0,
                #                  '_id': '52d1ce33c38a06510c000001',
                #                  'label': 'Moderator',
                #                  'rights': ['skip',
                #                             'queue-order',
                #                             'kick',
                #                             'ban',
                #                             'mute',
                #                             'set-dj',
                #                             'lock-queue',
                #                             'delete-chat',
                #                             'chat-mention'],
                #                  'type': 'mod'},
                #  'type': 'user-unsetrole',
                #  'user': {'__v': 0,
                #           '_force_updated': 1499443841892,
                #           '_id': '5628edc08d7d6a5600335d3d',
                #           'created': 1445522880666,
                #           'dubs': 0,
                #           'roleid': 1,
                #           'status': 1,
                #           'userInfo': {'__v': 0,
                #                        '_id': '5628edc08d7d6a5600335d3e',
                #                        'userid': '5628edc08d7d6a5600335d3d'},
                #           'username': 'iCel'}}
                modname = content['modUser']['username']
                modid = content['modUser']['_id']
                role = content['role_object']['label']
                roletype = content['role_object']['type']
                rights = content['role_object']['rights']
                username = content['user']['username']
                userid = content['user']['userInfo']['userid']
                logger_layer3.debug(
                    f'User {modname}#{modid} removed from role {role}/{roletype}({", ".join(rights)}) by {username}#'
                    f'{userid}')
            elif content_type.startswith('user_update'):
                # {'type': 'user_update_56096ce7a98a6b0300144e33',
                #  'user': {'_id': '5628b1c7e884391300d7427c',
                #           'updated': 1518781252893,
                #           'skippedCount': 0,
                #           'playedCount': 39994,
                #           'songsInQueue': 386,
                #           'active': True,
                #           'dubs': 15316,
                #           'order': 99999,
                #           'roomid': '561b1e59c90a9c0e00df610b',
                #           'userid': '56096ce7a98a6b0300144e33',
                #           '_user': '56096ce7a98a6b0300144e33',
                #           '__v': 0,
                #           'ot_token': None,
                #           'roleid': '5615fd84e596150061000003',
                #           'queuePaused': None,
                #           'authorized': True,
                #           'waitLine': 0}}
                user = content['user']
                userid = user['userid']
                skipped_count = user['skippedCount']
                played_count = user['playedCount']
                songs_in_queue = user['songsInQueue']
                dubs = user['dubs']
                logger_layer3.debug(
                    f'User updated {userid}, skip {skipped_count}, played {played_count}, queue {songs_in_queue}, '
                    f'dubs {dubs}')
            else:
                logger_layer3.info(
                    f'Received unknown message {content_type}: {pprint.pformat(content)}')
            yield content



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import asyncio
import inspect
import logging
import pprint
import re
import typing
from asyncio.events import AbstractEventLoop
from collections import Iterable, defaultdict
from inspect import iscoroutinefunction
from typing import List, Optional

from abot.cli import CommandCollection, Group
from abot.util import iterator_merge

logger = logging.getLogger(__name__)


class Abort(Exception):
    pass


class Backend:
    def configure(self, **config):
        raise NotImplementedError

    async def initialize(self):
        raise NotImplementedError()

    async def consume(self):
        raise NotImplementedError()

    def whoami(self) -> Optional['Entity']:
        raise NotImplementedError()

    def is_mentioned(self, message_event: 'MessageEvent') -> Optional[str]:
        username = self.whoami().username
        text = message_event.text
        if not (username and text):
            return None
        if len(text) < 2:
            return None
        pattern = r'^(@|!)?' + username + '(([,:]|\s)|$)'
        logger.debug(f'Checking if `{text}` matches {pattern}')
        if re.match(pattern, text):
            return username
        return None


class BotObject:
    @property
    def bot(self) -> 'Bot':
        if hasattr(self, '_bot'):
            return self._bot
        raise ValueError('Bot is not set in BotObject')

    @bot.setter
    def bot(self, bot: 'Bot'):
        if hasattr(self, '_bot'):
            raise ValueError(f'Bot {self._bot} is in place, cannot replace with {bot}')
        self._bot = bot

    @property
    def backend(self) -> Backend:
        raise NotImplementedError()


class Channel(BotObject):
    @property
    async def entities(self) -> List['Entity']:
        raise NotImplementedError()

    async def say(self, text: str):
        # Say something in the same channel as the message
        raise NotImplementedError()


class Entity(BotObject):
    @property
    def id(self) -> str:
        raise NotImplementedError()

    @property
    def username(self) -> str:
        raise NotImplementedError()

    async def tell(self, text: str):
        # Say something to the sender
        raise NotImplementedError()


class Event(BotObject):
    @property
    def sender(self) -> Entity:
        # Return the entity that sent this
        raise NotImplementedError()

    @property
    def channel(self) -> Channel:
        # Return the channel used to send the Event
        raise NotImplementedError()

    async def reply(self, text: str):
        # Reply to the message mentioning if possible
        raise NotImplementedError()


class MessageEvent(Event):
    @property
    def text(self) -> str:
        # Return the content of the message in plaintext
        raise NotImplementedError()


class Bot:
    def __init__(self):
        self.backends = {}
        self.event_handlers = defaultdict(set)
        self.message_handlers = set()
        self.forever_loop: asyncio.Future = None

    def attach_backend(self, backend: Backend):
        if backend in self.backends:
            raise ValueError(f'Backend {backend} is already attached to bot')
        iterator = self.backend_consume(backend)
        self.backends[backend] = iterator

    async def backend_consume(self, backend):
        while True:
            try:
                async for event in backend.consume():
                    yield event
            except Abort as e:
                logger.exception(f'Backend {backend} decided to abort bot execution')
                raise e from None
            except Exception:
                logger.exception(f'Exception in {backend} handled. Trying to recover.')

    def attach_command_group(self, group: Group):
        self.message_handlers.add(group)

    async def _handle_message(self, message):
        name = message.backend.is_mentioned(message)
        if not name:
            return
        cmd = CommandCollection(self.message_handlers)
        asyncio.ensure_future(cmd.async_message(message))

    def add_event_handler(self, event_class_or_func=None, *, func=None):
        if iscoroutinefunction(event_class_or_func):
            func = event_class_or_func
            event_class = None
        else:
            event_class = event_class_or_func

        if event_class and not isinstance(event_class, Iterable):
            event_class = (event_class,)

        def wrapper(f):
            nonlocal event_class
            assert asyncio.iscoroutinefunction(f), \
                f'Handler for {event_class} needs to be coroutine ({f})'
            if event_class is None:
                event_class = extract_possible_argument_types(f)

            for ec in event_class:
                self.event_handlers[f].add(ec)
            return f

        if func is None:
            return wrapper
        else:
            wrapper(func)

    async def _handle_event(self, event: Event):
        # Same event can be handled multiple times, Messages only once
        if isinstance(event, MessageEvent):
            logger.debug(f'Handling message {event.text}')
            await self._handle_message(event)
        runs = 0
        for cls in inspect.getmro(event.__class__):
            for handler, handled_classes in self.event_handlers.items():
                if cls not in handled_classes:
                    continue
                asyncio.ensure_future(self.run_event(handler, event))
                runs += 1
        if not runs:
            logger.debug(f'No message handler for {event}')

    async def _run_forever(self):
        continue_running = True

        for backend in self.backends:
            await backend.initialize()

        backend_iterators = {i: None for i in self.backends.values()}

        while continue_running:
            try:
                async for event in iterator_merge(backend_iterators):
                    await self._handle_event(event=event)
            except Abort as e:
                logger.info(f'Execution aborted by {e}')
                raise e from None
            except Exception as e:
                continue_running = await self.internal_exception_handler(e)

    async def run_forever(self):
        self.forever_loop = self._run_forever()
        return await self.forever_loop

    async def internal_exception_handler(self, exception):
        logger.error('Internal exception handled', exc_info=exception)

        tb = exception.__traceback__
        while tb.tb_next:
            tb = tb.tb_next

        logger.error(f'Locals: {pprint.pformat(tb.tb_frame.f_locals)}')
        return True

    async def run_event(self, func, event):
        logger.debug(f'Starting handling {event} with <{func.__name__}>')
        try:
            await func(event)
        except Abort as exception:
            self.forever_loop.set_exception(exception)
            logger.exception(f'Handling {event} in <{func.__name__}> aborted, stopping run')
        except Exception as exception:
            await self.handle_bot_exception(func, event, exception)
        else:
            logger.debug(f'Finished handling {event} with <{func.__name__}>')

    async def handle_bot_exception(self, func, event, exception):
        logger.exception(f'Failed running {event} in {func}')
        # await event.say(f':boom:... Houston we found a problem: ```{exception}```')

    def start(self, event_loop: AbstractEventLoop = None):
        return asyncio.ensure_future(self.run_forever(), loop=event_loop)


def extract_possible_argument_types(func) -> Iterable:
    args = typing.get_type_hints(func)
    if len(args) != 1:
        raise AttributeError(f'Function {func} can only have one argument if using type hinting')
    type_hint = next(iter(args.values()))
    if type_hint.__class__.__name__ in ('_Union', '_GenericAlias'):
        return type_hint.__args__
    return type_hint,



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import asyncio
import io
import shlex
from contextlib import contextmanager

import click
import click._unicodefun
import click.core
import click.utils
from click import *  # noqa

__all__ = click.__all__


@contextmanager
def stringio_wrapper(func):
    in_mem = io.StringIO()
    yield in_mem
    in_mem.seek(0)
    func(in_mem.read())


class Context(click.Context):
    async def async_invoke(*args, **kwargs):
        self, callback = args[:2]
        ctx = self

        # Deleted unused code from super class

        args = args[2:]
        with click.core.augment_usage_errors(self):
            with ctx:
                return await callback(*args, **kwargs)


class AsyncCommandMixin:
    def invoke(self, ctx):
        """Given a context, this invokes the attached callback (if it exists)
        in the right way.
        """
        if self.callback is not None:
            loop = asyncio.get_event_loop()
            return loop.run_until_complete(self.async_invoke(ctx))

    async def async_invoke(self, ctx):
        if self.callback is not None:
            return await ctx.async_invoke(self.callback, **ctx.params)

    def make_context(self, info_name, args, parent=None, **extra):
        for key, value in self.context_settings.items():
            if key not in extra:
                extra[key] = value
        ctx = Context(self, info_name=info_name, parent=parent, **extra)
        with ctx.scope(cleanup=False):
            self.parse_args(ctx, args)
        return ctx


class Command(AsyncCommandMixin, click.Command):
    pass


class AsyncMultiCommandMixin(AsyncCommandMixin):
    def invoke(self, ctx):
        args = ctx.protected_args + ctx.args
        _, cmd, _ = self.resolve_command(ctx, args)
        if ctx.__class__ != Context or isinstance(cmd, click.Command):
            click.MultiCommand.invoke(self, ctx)
        else:
            loop = asyncio.get_event_loop()
            return loop.run_until_complete(self.async_invoke(ctx))

    async def async_invoke(self, ctx):
        async def _process_result(value):
            if self.result_callback is not None:
                value = await ctx.async_invoke(self.result_callback, value,
                                               **ctx.params)
            return value

        if not ctx.protected_args:
            if self.invoke_without_command:
                if not self.chain:
                    return await Command.async_invoke(self, ctx)
                with ctx:
                    await Command.async_invoke(self, ctx)
                    return await _process_result([])
            ctx.fail('Missing command.')

        # Fetch args back out
        args = ctx.protected_args + ctx.args
        ctx.args = []
        ctx.protected_args = []

        if not self.chain:
            with ctx:
                cmd_name, cmd, args = self.resolve_command(ctx, args)
                ctx.invoked_subcommand = cmd_name
                await Command.async_invoke(self, ctx)
                sub_ctx = cmd.make_context(cmd_name, args, parent=ctx)
                with sub_ctx:
                    return await _process_result(await sub_ctx.command.async_invoke(sub_ctx))

        with ctx:
            ctx.invoked_subcommand = args and '*' or None
            await Command.async_invoke(self, ctx)

            contexts = []
            while args:
                cmd_name, cmd, args = self.resolve_command(ctx, args)
                sub_ctx = cmd.make_context(cmd_name, args, parent=ctx,
                                           allow_extra_args=True,
                                           allow_interspersed_args=False)
                contexts.append(sub_ctx)
                args, sub_ctx.args = sub_ctx.args, []

            rv = []
            for sub_ctx in contexts:
                with sub_ctx:
                    rv.append(await sub_ctx.command.async_invoke(sub_ctx))
            return _process_result(rv)


class MultiCommand(AsyncMultiCommandMixin, click.MultiCommand):
    pass


class AsyncGroupMixin(AsyncMultiCommandMixin):
    def command(self, *args, **kwargs):
        kwargs.setdefault('cls', Command)
        return super().command(*args, **kwargs)

    def group(self, *args, **kwargs):
        kwargs.setdefault('cls', Group)
        return super().command(*args, **kwargs)

    def invoke(self, ctx):
        return MultiCommand.invoke(self, ctx)

    async def async_invoke(self, ctx):
        return await MultiCommand.async_invoke(self, ctx)


class Group(AsyncGroupMixin, click.Group):
    pass


class AsyncCommandCollection(AsyncMultiCommandMixin):
    async def async_message(self, message: 'abot.bot.MessageEvent'):
        args = shlex.split(message.text)
        if not args:
            return
        prog_name = args.pop(0)
        try:
            try:
                with self.make_context(prog_name, args) as ctx:
                    await self.async_invoke(ctx)
            except (EOFError, KeyboardInterrupt):
                with stringio_wrapper(message.reply) as fd:
                    click.echo(file=fd, color=False)
                raise click.Abort()
            except click.ClickException as e:
                with stringio_wrapper(message.reply) as fd:
                    e.show(file=fd)
        except click.Abort:
            with stringio_wrapper(message.reply) as fd:
                click.echo('Aborted!', file=fd, color=False)


class CommandCollection(AsyncCommandCollection, click.CommandCollection):
    pass


def command(name=None, **attrs):
    attrs.setdefault('cls', Command)
    return click.command(name, **attrs)


def group(name=None, **attrs):
    attrs.setdefault('cls', Group)
    return click.group(name, **attrs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import asyncio
import logging
from typing import AsyncIterator, Dict, Optional

logger = logging.getLogger(__name__)


async def iterator_merge(iterators: Dict[AsyncIterator, Optional[asyncio.Future]]):
    while iterators:
        for iterator, value in list(iterators.items()):
            if not value:
                iterators[iterator] = asyncio.ensure_future(iterator.__anext__())

        tasks, _ = await asyncio.wait(iterators.values(), return_when=asyncio.FIRST_COMPLETED)
        for task in tasks:
            # We send the result up
            try:
                res = task.result()
                yield res
            except StopAsyncIteration:
                # We remove the task from the list
                for it, old_next in list(iterators.items()):
                    if task is old_next:
                        logger.debug(f'Iterator {it} finished consuming')
                        iterators.pop(it)
            else:
                # We remove the task from the key
                for it, old_next in list(iterators.items()):
                    if task is old_next:
                        iterators[it] = None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import asyncio
import json
import logging
from typing import List

import aiohttp
from aiohttp import WSMsgType
from aiohttp.formdata import FormData
from multidict import MultiDict

logger = logging.getLogger(__name__)


class SlackException(Exception):
    pass


class SlackCallException(SlackException):
    def __init__(self, message, method):
        self.method = method
        self.message = message
        super().__init__(message)

    def __str__(self):
        return f'Slack call {self.method}. {self.message}'


class SlackUseException(SlackException):
    pass


class SlackAPI:
    SLACK_RPC_PREFIX = 'https://slack.com/api/'
    SLACK_RTM_EVENTS = (
        'accounts_changed', 'bot_added', 'bot_changed', 'channel_archive', 'channel_created', 'channel_deleted',
        'channel_history_changed', 'channel_joined', 'channel_left', 'channel_marked', 'channel_rename',
        'channel_unarchive', 'commands_changed', 'dnd_updated', 'dnd_updated_user', 'email_domain_changed',
        'emoji_changed', 'file_change', 'file_comment_added', 'file_comment_deleted', 'file_comment_edited',
        'file_created', 'file_deleted', 'file_public', 'file_shared', 'file_unshared', 'goodbye', 'group_archive',
        'group_close', 'group_history_changed', 'group_joined', 'group_left', 'group_marked', 'group_open',
        'group_rename', 'group_unarchive', 'hello', 'im_close', 'im_created', 'im_history_changed', 'im_marked',
        'im_open', 'manual_presence_change', 'member_joined_channel', 'member_left_channel', 'message', 'pin_added',
        'pin_removed', 'pref_change', 'presence_change', 'reaction_added', 'reaction_removed', 'reconnect_url',
        'star_added', 'star_removed', 'subteam_created', 'subteam_self_added', 'subteam_self_removed',
        'subteam_updated', 'team_domain_change', 'team_join', 'team_migration_started', 'team_plan_change',
        'team_pref_change', 'team_profile_change', 'team_profile_delete', 'team_profile_reorder', 'team_rename',
        'user_change', 'user_typing',
    )
    SLACK_EVENTS_API_EVENTS = (
        'app_uninstalled', 'channel_archive', 'channel_created', 'channel_deleted', 'channel_history_changed',
        'channel_rename', 'channel_unarchive', 'dnd_updated', 'dnd_updated_user', 'email_domain_changed',
        'emoji_changed', 'file_change', 'file_comment_added', 'file_comment_deleted', 'file_comment_edited',
        'file_created', 'file_deleted', 'file_public', 'file_shared', 'file_unshared', 'grid_migration_finished',
        'grid_migration_started', 'group_archive', 'group_close', 'group_history_changed', 'group_open',
        'group_rename', 'group_unarchive', 'im_close', 'im_created', 'im_history_changed', 'im_open', 'link_shared',
        'member_joined_channel', 'member_left_channel', 'message', 'message.channels', 'message.groups', 'message.im',
        'message.mpim', 'pin_added', 'pin_removed', 'reaction_added', 'reaction_removed', 'star_added', 'star_removed',
        'subteam_created', 'subteam_members_changed', 'subteam_self_added', 'subteam_self_removed', 'subteam_updated',
        'team_domain_change', 'team_join', 'team_rename', 'tokens_revoked', 'url_verification', 'user_change'
    )

    def __init__(self, bot_token, event_loop=None):
        self.loop = event_loop or asyncio.get_event_loop()
        self.session = aiohttp.ClientSession(loop=self.loop)
        self.bot_token = bot_token
        self.groups = []
        self.users = []
        self.channels = []
        self.mpims = []
        self.ims = []
        self.bots = []
        self.ws_socket = None
        self.ws_ids = 1
        self.response_futures = {}

    async def request(self, method, url, data=None, headers=None):
        async with self.session.request(method=method, url=url, data=data, headers=headers) as response:
            return await response.json()

    async def call(self, method, **params):
        """
        Call an Slack Web API method

        :param method: Slack Web API method to call
        :param params: {str: object} parameters to method
        :return: dict()
        """
        url = self.SLACK_RPC_PREFIX + method
        data = FormData()
        data.add_fields(MultiDict(token=self.bot_token, charset='utf-8', **params))
        response_body = await self.request(
            method='POST',
            url=url,
            data=data
        )
        if 'warning' in response_body:
            logger.warning(f'Warnings received from API call {method}: {response_body["warning"]}')
        if 'ok' not in response_body:
            logger.error(f'No ok marker in slack API call {method} {params} => {response_body}')
            raise SlackCallException('There is no ok marker, ... strange', method=method)
        if not response_body['ok']:
            logger.error(f'Slack API call failed {method} {params} => {response_body}')
            raise SlackCallException(f'No OK response returned', method=method)
        return response_body

    async def slack_name_to_id(self, recipient):
        new_id = recipient
        if recipient.startswith('@'):
            username = recipient[1:]
            users = [user for user in self.users if user['bot_names'] == username and not user['deleted']]
            if len(users) > 1 or not users:
                logger.error(f'User {username} does not exist')
                raise SlackUseException(f'User {recipient} does not exist')
            new_id = users[0]['id']
        if recipient.startswith('#'):
            channel_name = recipient[1:]
            channels = [channel for channel in self.channels if
                        channel['bot_names'] == channel_name and not channel['archived']]
            if len(channels) > 1 or not channels:
                logger.error(f'Channel {channel_name} does not exist')
                raise SlackUseException(f'Channel {channel_name} does not exist')
            new_id = channels[0]['id']
        if new_id.startswith('U'):
            for im in self.ims:
                if im['user'] == new_id:
                    new_id = im['id']
                    break
            else:
                channel = await self.create_im(new_id)
                new_id = channel['channel']['id']
        return new_id

    async def userids_to_channel(self, userids):
        userids.sort()
        for mpim in self.mpims:
            if sorted(mpim['members']) == userids:
                return mpim['id']
        else:
            mpim = await self.create_mpim(users=userids)
            return mpim['id']

    def ws_send(self, body: dict):
        assert self.ws_socket and not self.ws_socket.closed, 'Writing to someone is only supported through ws'
        body['id'] = self.ws_ids
        self.ws_ids += 1
        logger.debug(f'Sending {body}')
        self.ws_socket.send_json(body)
        future = self.response_futures[body['id']] = asyncio.Future()
        return future

    async def write_to(self, recipients: List[str] or str, message: str):
        if not isinstance(recipients, str) and len(recipients) > 1:
            recipients_ids = await asyncio.gather(
                self.slack_name_to_id(recipient=recipient) for recipient in recipients
            )
            recipient = await self.userids_to_channel(userids=recipients_ids)
        else:
            recipient = recipients
        if recipient[0] in '@#U':  # User cannot be addressed directly, need to do it through DM channel
            recipient = await self.slack_name_to_id(recipient=recipient)
        assert recipient[0] in 'CDG', f'Programming error, receiver should start with (C|D|G) ({recipient}'
        return self.ws_send({
            'type': 'message',
            'channel': recipient,
            'text': message,
        })

    async def create_im(self, user: str):
        channel = await self.call('im.open', user=user, return_im=True)
        return channel

    async def create_mpim(self, users: List[str]):
        channel = await self.call('mpim.open', users=users)
        return channel

    def look_for_id(self, iterable, object_id):
        for item in iterable:
            if item['id'] == object_id:
                break
        else:
            item = None
        return item

    def ignore_message(self, message):
        logger.debug(f'Ignoring {message["type"]} message. {message}')
        return message

    def get_user_by_id(self, user_id):
        return self.look_for_id(self.users, user_id)

    handle_accounts_changed = ignore_message

    def handle_bot_added(self, message):
        bot_id = message['bot']['id']
        bot = self.look_for_id(self.bots, bot_id)
        if bot:
            logger.warning(f'Bot {bot_id} is to be added, but already exists, updating')
            bot.update(message['bot'])
        else:
            logger.debug(f'Adding bot {bot_id}')
            bot = {'deleted': False, 'updated': 0}
            bot.update(message['bot'])
            self.bots.append(bot)
        return message

    def handle_bot_changed(self, message):
        bot_id = message['bot']['id']
        bot = self.look_for_id(self.bots, bot_id)

        if bot:
            logger.debug(f'Bot {bot_id} changed')
            bot.update(message['bot'])
        else:
            logger.warning(f'Bot {bot_id} is to be changed, but does not exist, adding')
            bot = {'deleted': False, 'updated': 0}
            bot.update(message['bot'])
            self.bots.append(bot)
        return message

    def handle_channel_archive(self, message):
        channel_id = message['channel']
        channel = self.look_for_id(self.channels, channel_id)
        if channel:
            logger.debug(f'Channel {channel_id} has been archived. {message}')
            channel['is_archived'] = True
        else:
            logger.warning(f'Channel {channel_id} is not in the list of known channels, adding')
            self.channels.append({'id': channel_id, 'is_archived': True, "is_channel": True, })
        return message

    def handle_channel_created(self, message):
        channel_id = message['channel']['id']
        channel = self.look_for_id(self.channels, channel_id)
        if channel:
            logger.warning(f'Channel {channel_id} already exists, updating')
            channel.update(message['channel'])
        else:
            logger.debug(f'Channel {channel_id} has been created. {message["channel"]}')
            self.channels.append(dict(is_archived=False, is_channel=True, **message['channel']))
        return message

    def handle_channel_deleted(self, message):
        channel_id = message['channel']
        channel = self.look_for_id(self.channels, channel_id)
        if channel:
            logger.warning(f'Channel {channel_id} already exists, updating')
            channel.update(message['channel'])
        else:
            logger.debug(f'Channel {channel_id} has been created. {message["channel"]}')
            self.channels.append(dict(is_archived=False, is_channel=True, **message['channel']))
        return message

    handle_channel_history_changed = ignore_message

    def handle_channel_joined(self, message):
        channel_id = message['channel']['id']
        channel = self.look_for_id(self.channels, channel_id)
        if channel:
            logger.debug(f'Channel {channel_id} joined')
            channel.update(message['channel'])
        else:
            logger.warning(f'Joined previously unknown channel {channel_id}')
            self.channels.append(message['channel'])
        return message

    def handle_channel_left(self, message):
        channel_id = message['channel']
        channel = self.look_for_id(self.channels, channel_id)
        if channel:
            logger.debug(f'Left channel {channel_id}')
            channel['is_member'] = False
        else:
            logger.warning(f'Left previously unknown channel {channel_id}')
            self.channels.append(dict(id=channel_id, is_channel=True, ))
        return message

    def handle_channel_marked(self, message):
        channel_id = message['channel']
        channel = self.look_for_id(self.channels, channel_id)
        if channel:
            logger.debug(f'Channel mark event for {channel_id}, doing nothing')
        else:
            logger.warning(f'Mark on previously unknown channel {channel_id}')
            self.channels.append(dict(id=channel_id, is_channel=True, ))
        return message

    def handle_channel_rename(self, message):
        channel_id = message['channel']['id']
        channel = self.look_for_id(self.channels, channel_id)
        if channel:
            logger.debug(f'Channel {channel_id} renamed')
            channel.update(message['channel'])
        else:
            logger.warning(f'Rename of previously unknown channel {channel_id}')
            self.channels.append(dict(is_channel=True, **message['channel']))
        return message

    def handle_channel_unarchive(self, message):
        channel_id = message['channel']
        channel = self.look_for_id(self.channels, channel_id)
        if channel:
            logger.debug(f'Channel {channel_id} has been unarchived. {message}')
            channel['is_archived'] = False
        else:
            logger.warning(f'Channel {channel_id} is not in the list of known channels, unarchiving')
            self.channels.append({'id': channel_id, 'is_archived': False, "is_channel": True, })
        return message

    handle_commands_changed = ignore_message
    handle_dnd_updated = ignore_message
    handle_dnd_updated_user = ignore_message
    handle_email_domain_changed = ignore_message
    handle_emoji_changed = ignore_message
    handle_file_change = ignore_message
    handle_file_comment_added = ignore_message
    handle_file_comment_deleted = ignore_message
    handle_file_comment_edited = ignore_message
    handle_file_created = ignore_message
    handle_file_deleted = ignore_message
    handle_file_public = ignore_message
    handle_file_shared = ignore_message
    handle_file_unshared = ignore_message
    handle_goodbye = ignore_message  # TODO: Needs to be handled
    handle_grid_migration_finished = ignore_message
    handle_grid_migration_started = ignore_message

    def handle_group_archive(self, message):
        group_id = message['channel']
        group = self.look_for_id(self.groups, group_id)
        if group:
            logger.debug(f'Group {group_id} has been archived. {message}')
            group['is_archived'] = True
        else:
            logger.warning(f'Group {group_id} is not in the list of known groups, archiving')
            self.channels.append({'id': group_id, 'is_archived': True, "is_group": True, })
        return message

    def handle_group_close(self, message):
        group_id = message['channel']
        group = self.look_for_id(self.groups, group_id)
        if group:
            logger.debug(f'Marking group {group_id} as closed. {message}')
            group['is_open'] = False
        else:
            logger.warning(f'Marking non existent group as closed. {message}')
            self.groups.append({'is_group': True, 'id': group_id, 'is_open': False})
        return message

    handle_group_history_changed = ignore_message

    def handle_group_joined(self, message):
        group_id = message['channel']['id']
        group = self.look_for_id(self.groups, group_id)
        if group:
            logger.warning(f'Creating already existing group {group_id}. {message}')
            group.update(message['channel'])
        else:
            logger.debug(f'Joined group {group_id}. {message}')
            self.groups.append(message['channel'])
        return message

    def handle_group_left(self, message):
        group_id = message['channel']
        group = self.look_for_id(self.groups, group_id)
        if group:
            logger.debug(f'Left group {group_id}')
            group['is_member'] = False
        else:
            logger.warning(f'Left previously unknown group {group_id}')
            self.groups.append(dict(id=group_id, is_group=True, ))
        return message

    def handle_group_marked(self, message):
        group_id = message['channel']
        group = self.look_for_id(self.groups, group_id)
        if group:
            logger.debug(f'Channel mark event for {group_id}, doing nothing')
        else:
            logger.warning(f'Mark on previously unknown group {group_id}')
            self.groups.append(dict(id=group_id, is_group=True, ))
        return message

    def handle_group_open(self, message):
        group_id = message['channel']
        group = self.look_for_id(self.groups, group_id)
        if group:
            logger.debug(f'Group {group_id} open')
            group['is_open'] = True
        else:
            logger.warning(f'Open previously unknown group {group_id}')
            self.groups.append(dict(id=group_id, is_group=True, is_open=True))
        return message

    def handle_group_rename(self, message):
        group_id = message['channel']['id']
        group = self.look_for_id(self.groups, group_id)
        if group:
            logger.debug(f'Group {group_id} rename')
            group['name'] = message['channel']['name']
        else:
            logger.warning(f'Rename previously unknown group {group_id}')
            self.groups.append(dict(id=group_id, is_group=True, name=message['channel']['name']))
        return message

    def handle_group_unarchive(self, message):
        group_id = message['channel']['id']
        group = self.look_for_id(self.groups, group_id)
        if group:
            logger.debug(f'Marking group {group_id} unarchived')
            group['is_archived'] = False
        else:
            logger.warning(f'Unarchiving previously unknown group {group_id}')
            self.groups.append(dict(id=group_id, is_group=True, is_archived=False))
        return message

    def handle_hello(self, message):
        logger.info('Correctly connected to RTM stream')
        return message

    def handle_im_close(self, message):
        im_id = message['channel']
        im = self.look_for_id(self.ims, im_id)
        if im:
            logger.debug(f'Marking im {im_id} as closed. {message}')
            im['is_open'] = False
        else:
            logger.warning(f'Marking non existent im as closed. {message}')
            self.ims.append({'is_im': True, 'id': im_id, 'is_open': False})
        return message

    def handle_im_created(self, message):
        im_id = message['channel']['id']
        im = self.look_for_id(self.ims, im_id)
        if im:
            logger.warning(f'Channel {im_id} already exists, updating')
            im.update(message['im'])
        else:
            logger.debug(f'Channel {im_id} has been created. {message}')
            self.ims.append(dict(is_archived=False, is_im=True, **message['channel']))
        return message

    handle_im_history_changed = ignore_message
    handle_im_marked = ignore_message

    def handle_im_open(self, message):
        im_id = message['channel']
        im = self.look_for_id(self.ims, im_id)
        if im:
            logger.debug(f'Marking im {im_id} as closed. {message}')
            im['is_open'] = True
        else:
            logger.warning(f'Marking non existent im as closed. {message}')
            self.ims.append({'is_im': True, 'id': im_id, 'is_open': True})
        return message

    def handle_manual_presence_change(self, message):
        user_id = message['user']
        user = self.look_for_id(self.users, user_id)
        presence = message["presence"]
        if user:
            logger.debug(f'User {user_id} presence manually updated to {presence}')
            user['presence'] = presence
        else:
            logger.warning(f'Setting presence for previously unknown user {user_id}')
            self.users.append(dict(id=user_id, presence=presence))
        return message

    def handle_member_joined_channel(self, message):
        channel_id, channel_type = message['channel'], message['channel_type']
        if channel_type == 'C':
            channel = self.look_for_id(self.channels, channel_id)
        elif channel_type == 'G':
            channel = self.look_for_id(self.groups, channel_id)
        else:
            logger.warning(f'Unknown type of channel type {channel_type}, ignoring')
            return message

        user = message['user']
        if channel:
            if 'members' not in channel:
                logger.debug(f'No previously gathered members for channel {channel_id}, adding {user}')
                channel['members'] = [user]
            elif user not in channel['members']:
                logger.debug(f'Adding {user} to members')
                channel['members'].append(user)
            else:
                logger.warning(f'User {user} is already part of the members of {channel_id}')
        else:
            logger.warning(f'Adding previously unknown channel/group {channel_id} the member {user}')
            if channel_type == 'C':
                self.channels.append({'id': channel_id, 'is_channel': True, 'members': [user]})
            elif channel_type == 'G':
                self.groups.append({'id': channel_id, 'is_group': True, 'members': [user]})

        return message

    def handle_member_left_channel(self, message):
        channel_id, channel_type = message['channel'], message['channel_type']
        if channel_type == 'C':
            channel = self.look_for_id(self.channels, channel_id)
        elif channel_type == 'G':
            channel = self.look_for_id(self.groups, channel_id)
        else:
            logger.warning(f'Unknown type of channel type {channel_type}, ignoring')
            return message

        user = message['user']
        if channel:
            if 'members' not in channel:
                logger.debug(f'No previously gathered members for channel {channel_id}, creating empty (no {user})')
                channel['members'] = []
            elif user in channel['members']:
                logger.debug(f'Adding {user} to members')
                channel['members'].remove(user)
            else:
                logger.warning(f'User {user} is already not part of the members of {channel_id}')
        else:
            logger.warning(f'Adding previously unknown channel/group {channel_id} with empty members (no {user})')
            if channel_type == 'C':
                self.channels.append({'id': channel_id, 'is_channel': True, 'members': []})
            elif channel_type == 'G':
                self.groups.append({'id': channel_id, 'is_group': True, 'members': []})

        return message

    handle_pin_added = ignore_message
    handle_pin_removed = ignore_message
    handle_pref_change = ignore_message

    def handle_presence_change(self, message):
        user_id = message['user']
        user = self.look_for_id(self.users, user_id)
        presence = message["presence"]
        if user:
            logger.debug(f'User {user_id} presence updated to {presence}')
            user['presence'] = presence
        else:
            logger.warning(f'Setting presence for previously unknown user {user_id}')
            self.users.append(dict(id=user_id, presence=presence))
        return message

    handle_reaction_added = ignore_message
    handle_reaction_removed = ignore_message
    handle_reconnect_url = ignore_message
    handle_star_added = ignore_message
    handle_star_removed = ignore_message
    handle_subteam_created = ignore_message
    handle_subteam_members_changed = ignore_message
    handle_subteam_self_added = ignore_message
    handle_subteam_self_removed = ignore_message
    handle_subteam_updated = ignore_message
    handle_team_domain_change = ignore_message

    def handle_team_join(self, message):
        user_id = message['user']['id']
        user = self.look_for_id(self.users, user_id)

        if user:
            logger.warning(f'User {user} that was just created already existed. {message}')
            user.update(message['user'])
        else:
            logger.debug(f'Adding user {user_id} changed. {message}')
            self.users.append(message['user'])
        return message

    handle_team_migration_started = ignore_message
    handle_team_plan_change = ignore_message
    handle_team_pref_change = ignore_message
    handle_team_profile_change = ignore_message
    handle_team_profile_delete = ignore_message
    handle_team_profile_reorder = ignore_message
    handle_team_rename = ignore_message

    def handle_user_change(self, message):
        user_id = message['user']['id']
        user = self.look_for_id(self.users, user_id)

        if user:
            logger.debug(f'User {user} updated. {message}')
            user.update(message['user'])
        else:
            logger.warning(f'Previously non existent user {user_id} changed. {message}')
            self.users.append(message['user'])
        return message

    handle_user_typing = ignore_message

    def rtm_handler(self, ws_message):
        """
        Handle a message, processing it internally if required. If it's a message that should go outside the bot,
        this function will return True

        :param message:
        :return: Boolean if message should be yielded
        """
        message = json.loads(ws_message.data)
        if 'reply_to' in message:
            reply_to = message['reply_to']
            future = self.response_futures.pop(reply_to, None)
            if future is None:
                logger.error(f'This should not happen, received reply to unknown message! {message}')
                return None
            future.set_result(message)
            return None
        if 'type' not in message:
            logger.error(f'No idea what this could be {message}')
            return
        message_type = message['type']

        if hasattr(self, f'handle_{message_type}'):
            function = getattr(self, f'handle_{message_type}')
            return function(message)

        if message_type in self.SLACK_RTM_EVENTS:
            logger.debug(f'Unhandled {message_type}. {message}')
        else:
            logger.warning(f'Unknown {message_type}. {message}')
        return message

    async def rtm_api_consume(self):
        response = await self.call('rtm.start', simple_latest=False, no_unreads=False, mpim_aware=True)
        self.channels = response['channels']
        self.groups = response['groups']
        self.ims = response['ims']
        self.mpims = response['mpims']
        self.users = response['users']
        self.bots = response['bots']
        logger.debug(f'Connect url {response["url"]}')
        async with self.session.ws_connect(url=response['url']) as self.ws_socket:
            async for ws_message in self.ws_socket:
                if ws_message.tp == WSMsgType.text:
                    message_content = self.rtm_handler(ws_message=ws_message)
                    if message_content:
                        yield message_content
                elif ws_message.tp in (WSMsgType.closed, WSMsgType.error):
                    logger.info('Finishing ws, %s', ws_message)
                    if not self.ws_socket.closed:
                        await self.ws_socket.close()
                    break

    def __del__(self):
        asyncio.ensure_future(self.session.close(), loop=self.loop)
        self.session.close()


for method in dir(SlackAPI):
    all_events = set()
    all_events.update(set(SlackAPI.SLACK_RTM_EVENTS))
    all_events.update(set(SlackAPI.SLACK_EVENTS_API_EVENTS))
    if method.startswith('handle_'):
        handle, event_name = method.split('_', 1)
        assert event_name in all_events, f'SlackAPI defines unregistered handler {method}'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import subprocess

from setuptools import find_packages, setup

import abot


def git_remote_tags():
    tags = subprocess.check_output('git ls-remote git://github.com/txomon/abot --tags v*'.split()).decode()
    valid_tags = {}
    for tag in tags.splitlines():
        # 8626062a1526fdf239fd52d35d91cf7896729e8f        refs/tags/test^{}
        if tag.endswith('^{}'):
            splits = [p for p in tag.split() if p]
            commit, tagname = splits
            tagname = tagname[len('refs/tags/'):-len('^{}')]
            valid_tags[commit] = tagname

    return valid_tags


def git_remote_commits():
    subprocess.check_output('git fetch git://github.com/txomon/abot master:refs/reference'.split()).decode()
    git_log = subprocess.check_output('git log --format=%H refs/reference'.split()).decode()
    return git_log.splitlines()


def git_get_closest_commit(valid_tags, remote_commits):
    git_log = subprocess.check_output('git log --format=%H'.split()).decode()
    common_commit = None
    common_to_current = None
    current_commit = None
    for number, commit in enumerate(git_log.splitlines()):
        if not current_commit:
            current_commit = commit
        if not common_commit and commit in remote_commits:
            common_commit = commit
            common_to_current = number
        if commit in valid_tags:
            return {
                'tag': valid_tags[commit],
                'tag_to_current': number,
                'commit': current_commit,
                'common_commit': common_commit,
                'common_to_current': common_to_current,
            }


def get_version():
    # Badass revision v1.2.9a1.dev40+20.ab31d12b-dirty
    #                 ********######@@@@@@@@@@@@++++++
    #          ----------/  ----/        /         \---
    #         /            /            /              \
    # {version_string}{dev_string}{local_string}{status_string}
    try:
        dev_string = local_string = status_string = ''
        # Get dirtiness
        valid_tags = git_remote_tags()
        remote_commits = git_remote_commits()
        info = git_get_closest_commit(valid_tags, remote_commits)
        if not info:  # If there is no match, just go to offline mode
            raise Exception()

        # First, public release string (v1.2.12a3) AKA version_string
        version_string = info["tag"]
        code_version = f'v{abot.__version__}'
        if version_string != code_version:
            raise TypeError(f'ABot version in abot/__init__.py needs to be updated to {version_string}')

        common_to_current = info['common_to_current']
        if common_to_current is None:
            raise TypeError(f'This repo and upstream are messy, no common commits. Report this situation!')

        tag_to_current = info['tag_to_current']
        if common_to_current > tag_to_current:
            raise TypeError(f'This is an impossible situation, no tag can exist in uncommon path')
        tag_to_common = tag_to_current - common_to_current

        # Second, official non-public dev release string (.dev30) AKA dev_string
        if tag_to_common != 0:
            dev_string = f'.dev{tag_to_common}'

        # Third, non-official non-public user release number (+29-ab38ecd99)
        if common_to_current:
            local_string = f'+{common_to_current}.{info["commit"][:8]}'

        # Fourth, dirtiness flag
        is_dirty = bool(subprocess.check_output('git status -s'.split()).decode())
        if is_dirty:
            status_string = '.dirty'

        return f'{version_string}{dev_string}{local_string}{status_string}'
    except TypeError:
        raise
    except Exception:
        version = abot.__version__
        return f'v{version}+unknown'


setup(
    name='abot',
    version=get_version(),
    description='Bot creation library',
    long_description=open('README.rst').read(),
    url='https://github.com/txomon/abot',
    author='Javier Domingo Cansino',
    author_email='javierdo1@gmail.com',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(exclude=['tests']),
    license='MIT',
    python_requires='>=3.6',
    include_package_data=True,
    zip_safe=False,
    keywords=['slack', 'dubtrack', 'bot', 'async', 'asyncio'],
    install_requires=[
        'aiohttp',
        'click',
    ],
)
